# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
#
# Authors: Jean-Emile DARTOIS <jean-emile.dartois@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

from oslo_log import log
from oslo_utils import timeutils

from watcher._i18n import _
from watcher.common import clients
from watcher.common import exception
from watcher.decision_engine.datasources import base


LOG = log.getLogger(__name__)


try:
    from ceilometerclient import exc
    HAS_CEILCLIENT = True
except ImportError:
    HAS_CEILCLIENT = False


class CeilometerHelper(base.DataSourceBase):

    NAME = 'ceilometer'
    METRIC_MAP = dict(host_cpu_usage='compute.node.cpu.percent',
                      host_ram_usage='hardware.memory.used',
                      host_outlet_temp='hardware.ipmi.node.outlet_temperature',
                      host_inlet_temp='hardware.ipmi.node.temperature',
                      host_airflow='hardware.ipmi.node.airflow',
                      host_power='hardware.ipmi.node.power',
                      instance_cpu_usage='cpu_util',
                      instance_ram_usage='memory.resident',
                      instance_ram_allocated='memory',
                      instance_l3_cache_usage='cpu_l3_cache',
                      instance_root_disk_size='disk.root.size',
                      )

    def __init__(self, osc=None):
        """:param osc: an OpenStackClients instance"""
        self.osc = osc if osc else clients.OpenStackClients()
        self.ceilometer = self.osc.ceilometer()
        LOG.warning("Ceilometer API is deprecated and Ceilometer Datasource "
                    "module is no longer maintained. We recommend to use "
                    "Gnocchi instead.")

    @staticmethod
    def format_query(user_id, tenant_id, resource_id,
                     user_ids, tenant_ids, resource_ids):
        query = []

        def query_append(query, _id, _ids, field):
            if _id:
                _ids = [_id]
            for x_id in _ids:
                query.append({"field": field, "op": "eq", "value": x_id})

        query_append(query, user_id, (user_ids or []), "user_id")
        query_append(query, tenant_id, (tenant_ids or []), "project_id")
        query_append(query, resource_id, (resource_ids or []), "resource_id")

        return query

    def _timestamps(self, start_time, end_time):

        def _format_timestamp(_time):
            if _time:
                if isinstance(_time, datetime.datetime):
                    return _time.isoformat()
                return _time
            return None

        start_timestamp = _format_timestamp(start_time)
        end_timestamp = _format_timestamp(end_time)

        if ((start_timestamp is not None) and (end_timestamp is not None) and
                (timeutils.parse_isotime(start_timestamp) >
                 timeutils.parse_isotime(end_timestamp))):
            raise exception.Invalid(
                _("Invalid query: %(start_time)s > %(end_time)s") % dict(
                    start_time=start_timestamp, end_time=end_timestamp))
        return start_timestamp, end_timestamp

    def build_query(self, user_id=None, tenant_id=None, resource_id=None,
                    user_ids=None, tenant_ids=None, resource_ids=None,
                    start_time=None, end_time=None):
        """Returns query built from given parameters.

        This query can be then used for querying resources, meters and
        statistics.
        :param user_id: user_id, has a priority over list of ids
        :param tenant_id: tenant_id, has a priority over list of ids
        :param resource_id: resource_id, has a priority over list of ids
        :param user_ids: list of user_ids
        :param tenant_ids: list of tenant_ids
        :param resource_ids: list of resource_ids
        :param start_time: datetime from which measurements should be collected
        :param end_time: datetime until which measurements should be collected
        """

        query = self.format_query(user_id, tenant_id, resource_id,
                                  user_ids, tenant_ids, resource_ids)

        start_timestamp, end_timestamp = self._timestamps(start_time,
                                                          end_time)

        if start_timestamp:
            query.append({"field": "timestamp", "op": "ge",
                          "value": start_timestamp})
        if end_timestamp:
            query.append({"field": "timestamp", "op": "le",
                          "value": end_timestamp})
        return query

    def query_retry_reset(self, exception_instance):
        if isinstance(exception_instance, exc.HTTPUnauthorized):
            self.osc.reset_clients()
            self.ceilometer = self.osc.ceilometer()

    def list_metrics(self):
        """List the user's meters."""
        meters = self.query_retry(f=self.ceilometer.meters.list)
        if not meters:
            return set()
        else:
            return meters

    def check_availability(self):
        status = self.query_retry(self.ceilometer.resources.list)
        if status:
            return 'available'
        else:
            return 'not available'

    def query_sample(self, meter_name, query, limit=1):
        return self.query_retry(f=self.ceilometer.samples.list,
                                meter_name=meter_name,
                                limit=limit,
                                q=query)

    def statistic_aggregation(self, resource=None, resource_type=None,
                              meter_name=None, period=300, granularity=300,
                              aggregate='mean'):
        end_time = datetime.datetime.utcnow()
        start_time = end_time - datetime.timedelta(seconds=int(period))

        meter = self._get_meter(meter_name)

        if aggregate == 'mean':
            aggregate = 'avg'
        elif aggregate == 'count':
            aggregate = 'avg'
            LOG.warning('aggregate type count not supported by ceilometer,'
                        ' replaced with mean.')

        resource_id = resource.uuid
        if resource_type == 'compute_node':
            resource_id = "%s_%s" % (resource.hostname, resource.hostname)

        query = self.build_query(
            resource_id=resource_id, start_time=start_time, end_time=end_time)
        statistic = self.query_retry(f=self.ceilometer.statistics.list,
                                     meter_name=meter,
                                     q=query,
                                     period=period,
                                     aggregates=[
                                         {'func': aggregate}])

        item_value = None
        if statistic:
            item_value = statistic[-1]._info.get('aggregate').get(aggregate)
            if meter_name == 'host_airflow':
                # Airflow from hardware.ipmi.node.airflow is reported as
                # 1/10 th of actual CFM
                item_value *= 10
        return item_value

    def statistic_series(self, resource=None, resource_type=None,
                         meter_name=None, start_time=None, end_time=None,
                         granularity=300):
        raise NotImplementedError(
            _('Ceilometer helper does not support statistic series method'))

    def get_host_cpu_usage(self, resource, period,
                           aggregate, granularity=None):

        return self.statistic_aggregation(
            resource, 'compute_node', 'host_cpu_usage', period,
            aggregate, granularity)

    def get_host_ram_usage(self, resource, period,
                           aggregate, granularity=None):

        return self.statistic_aggregation(
            resource, 'compute_node', 'host_ram_usage', period,
            aggregate, granularity)

    def get_host_outlet_temp(self, resource, period,
                             aggregate, granularity=None):

        return self.statistic_aggregation(
            resource, 'compute_node', 'host_outlet_temp', period,
            aggregate, granularity)

    def get_host_inlet_temp(self, resource, period,
                            aggregate, granularity=None):

        return self.statistic_aggregation(
            resource, 'compute_node', 'host_inlet_temp', period,
            aggregate, granularity)

    def get_host_airflow(self, resource, period,
                         aggregate, granularity=None):

        return self.statistic_aggregation(
            resource, 'compute_node', 'host_airflow', period,
            aggregate, granularity)

    def get_host_power(self, resource, period,
                       aggregate, granularity=None):

        return self.statistic_aggregation(
            resource, 'compute_node', 'host_power', period,
            aggregate, granularity)

    def get_instance_cpu_usage(self, resource, period,
                               aggregate, granularity=None):

        return self.statistic_aggregation(
            resource, 'instance', 'instance_cpu_usage', period,
            aggregate, granularity)

    def get_instance_ram_usage(self, resource, period,
                               aggregate, granularity=None):

        return self.statistic_aggregation(
            resource, 'instance', 'instance_ram_usage', period,
            aggregate, granularity)

    def get_instance_ram_allocated(self, resource, period,
                                   aggregate, granularity=None):

        return self.statistic_aggregation(
            resource, 'instance', 'instance_ram_allocated', period,
            aggregate, granularity)

    def get_instance_l3_cache_usage(self, resource, period,
                                    aggregate, granularity=None):

        return self.statistic_aggregation(
            resource, 'instance', 'instance_l3_cache_usage', period,
            aggregate, granularity)

    def get_instance_root_disk_size(self, resource, period,
                                    aggregate, granularity=None):

        return self.statistic_aggregation(
            resource, 'instance', 'instance_root_disk_size', period,
            aggregate, granularity)



# -*- encoding: utf-8 -*-
# Copyright 2017 NEC Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log

from watcher.common import cinder_helper
from watcher.common import exception
from watcher.decision_engine.model.collector import base
from watcher.decision_engine.model import element
from watcher.decision_engine.model import model_root
from watcher.decision_engine.model.notification import cinder
from watcher.decision_engine.scope import storage as storage_scope

LOG = log.getLogger(__name__)


class CinderClusterDataModelCollector(base.BaseClusterDataModelCollector):
    """Cinder cluster data model collector

    The Cinder cluster data model collector creates an in-memory
    representation of the resources exposed by the storage service.
    """
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "availability_zones": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "name": {
                                "type": "string"
                            }
                        },
                        "additionalProperties": False
                    }
                },
                "volume_types": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "name": {
                                "type": "string"
                            }
                        },
                        "additionalProperties": False
                    }
                },
                "exclude": {
                    "type": "array",
                    "items": {
                        "type": "object",
                        "properties": {
                            "storage_pools": {
                                "type": "array",
                                "items": {
                                    "type": "object",
                                    "properties": {
                                        "name": {
                                            "type": "string"
                                        }
                                    },
                                    "additionalProperties": False
                                }
                            },
                            "volumes": {
                                "type": "array",
                                "items": {
                                    "type": "object",
                                    "properties": {
                                        "uuid": {
                                            "type": "string"
                                        }
                                    },
                                    "additionalProperties": False
                                }
                            },
                            "projects": {
                                "type": "array",
                                "items": {
                                    "type": "object",
                                    "properties": {
                                        "uuid": {
                                            "type": "string"
                                        }
                                    },
                                    "additionalProperties": False
                                }
                            },
                            "additionalProperties": False
                        }
                    }
                }
            },
            "additionalProperties": False
        }
    }

    def __init__(self, config, osc=None):
        super(CinderClusterDataModelCollector, self).__init__(config, osc)

    @property
    def notification_endpoints(self):
        """Associated notification endpoints

        :return: Associated notification endpoints
        :rtype: List of :py:class:`~.EventsNotificationEndpoint` instances
        """
        return [
            cinder.CapacityNotificationEndpoint(self),
            cinder.VolumeCreateEnd(self),
            cinder.VolumeDeleteEnd(self),
            cinder.VolumeUpdateEnd(self),
            cinder.VolumeAttachEnd(self),
            cinder.VolumeDetachEnd(self),
            cinder.VolumeResizeEnd(self)
        ]

    def get_audit_scope_handler(self, audit_scope):
        self._audit_scope_handler = storage_scope.StorageScope(
            audit_scope, self.config)
        if self._data_model_scope is None or (
            len(self._data_model_scope) > 0 and (
                self._data_model_scope != audit_scope)):
            self._data_model_scope = audit_scope
            self._cluster_data_model = None
        LOG.debug("audit scope %s", audit_scope)
        return self._audit_scope_handler

    def execute(self):
        """Build the storage cluster data model"""
        LOG.debug("Building latest Cinder cluster data model")

        if self._audit_scope_handler is None:
            LOG.debug("No audit, Don't Build storage data model")
            return
        if self._data_model_scope is None:
            LOG.debug("No audit scope, Don't Build storage data model")
            return

        builder = CinderModelBuilder(self.osc)
        return builder.execute(self._data_model_scope)


class CinderModelBuilder(base.BaseModelBuilder):
    """Build the graph-based model

    This model builder adds the following data"
    - Storage-related knowledge (Cinder)

    """
    def __init__(self, osc):
        self.osc = osc
        self.model = model_root.StorageModelRoot()
        self.cinder = osc.cinder()
        self.cinder_helper = cinder_helper.CinderHelper(osc=self.osc)

    def _add_physical_layer(self):
        """Add the physical layer of the graph.

        This includes components which represent actual infrastructure
        hardware.
        """
        for snode in self.call_retry(
                self.cinder_helper.get_storage_node_list):
            self.add_storage_node(snode)
        for pool in self.call_retry(self.cinder_helper.get_storage_pool_list):
            pool = self._build_storage_pool(pool)
            self.model.add_pool(pool)
            storage_name = getattr(pool, 'name')
            try:
                storage_node = self.model.get_node_by_name(
                    storage_name)
                # Connect the instance to its compute node
                self.model.map_pool(pool, storage_node)
            except exception.StorageNodeNotFound:
                continue

    def add_storage_node(self, node):
        # Build and add base node.
        storage_node = self.build_storage_node(node)
        self.model.add_node(storage_node)

    def add_storage_pool(self, pool):
        storage_pool = self._build_storage_pool(pool)
        self.model.add_pool(storage_pool)

    def build_storage_node(self, node):
        """Build a storage node from a Cinder storage node

        :param node: A storage node
        :type node: :py:class:`~cinderclient.v3.services.Service`
        """
        # node.host is formatted as host@backendname since ocata,
        # or may be only host as of ocata
        backend = ""
        try:
            backend = node.host.split('@')[1]
        except IndexError:
            pass

        volume_type = self.call_retry(
            self.cinder_helper.get_volume_type_by_backendname, backend)

        # build up the storage node.
        node_attributes = {
            "host": node.host,
            "zone": node.zone,
            "state": node.state,
            "status": node.status,
            "volume_type": volume_type}

        storage_node = element.StorageNode(**node_attributes)
        return storage_node

    def _build_storage_pool(self, pool):
        """Build a storage pool from a Cinder storage pool

        :param pool: A storage pool
        :type pool: :py:class:`~cinderclient.v3.pools.Pool`
        :raises: exception.InvalidPoolAttributeValue
        """
        # build up the storage pool.

        attrs = ["total_volumes", "total_capacity_gb",
                 "free_capacity_gb", "provisioned_capacity_gb",
                 "allocated_capacity_gb"]

        node_attributes = {"name": pool.name}
        for attr in attrs:
            try:
                node_attributes[attr] = int(getattr(pool, attr))
            except AttributeError:
                LOG.debug("Attribute %s for pool %s is not provided",
                          attr, pool.name)
            except ValueError:
                raise exception.InvalidPoolAttributeValue(
                    name=pool.name, attribute=attr)

        storage_pool = element.Pool(**node_attributes)
        return storage_pool

    def _add_virtual_layer(self):
        """Add the virtual layer to the graph.

        This layer is the virtual components of the infrastructure.
        """
        self._add_virtual_storage()

    def _add_virtual_storage(self):
        volumes = self.call_retry(self.cinder_helper.get_volume_list)
        for vol in volumes:
            volume = self._build_volume_node(vol)
            self.model.add_volume(volume)
            pool_name = getattr(vol, 'os-vol-host-attr:host')
            if pool_name is None:
                # The volume is not attached to any pool
                continue
            try:
                pool = self.model.get_pool_by_pool_name(
                    pool_name)
                self.model.map_volume(volume, pool)
            except exception.PoolNotFound:
                continue

    def _build_volume_node(self, volume):
        """Build an volume node

        Create an volume node for the graph using cinder and the
        `volume` cinder object.
        :param instance: Cinder Volume object.
        :return: A volume node for the graph.
        """
        attachments = [{k: v for k, v in iter(d.items()) if k in (
            'server_id', 'attachment_id')} for d in volume.attachments]

        volume_attributes = {
            "uuid": volume.id,
            "size": volume.size,
            "status": volume.status,
            "attachments": attachments,
            "name": volume.name or "",
            "multiattach": volume.multiattach,
            "snapshot_id": volume.snapshot_id or "",
            "project_id": getattr(volume, 'os-vol-tenant-attr:tenant_id'),
            "metadata": volume.metadata,
            "bootable": volume.bootable}

        return element.Volume(**volume_attributes)

    def execute(self, model_scope):
        """Instantiates the graph with the openstack cluster data.

        The graph is populated along 2 layers: virtual and physical. As each
        new layer is built connections are made back to previous layers.
        """
        # TODO(Dantali0n): Use scope to limit size of model
        self._add_physical_layer()
        self._add_virtual_layer()
        return self.model



# -*- encoding: utf-8 -*-
# Copyright (c) 2019 European Organization for Nuclear Research (CERN)
#
# Authors: Corne Lukken <info@dantalion.nl>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log
from oslo_serialization import jsonutils

from watcher.common import exception
from watcher.decision_engine.datasources.grafana_translator.base import \
    BaseGrafanaTranslator

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class InfluxDBGrafanaTranslator(BaseGrafanaTranslator):
    """Grafana translator to communicate with InfluxDB database"""

    NAME = 'influxdb'

    def __init__(self, data):
        super(InfluxDBGrafanaTranslator, self).__init__(data)

    def build_params(self):
        """"""

        data = self._data

        retention_period = None
        available_periods = CONF.grafana_translators.retention_periods.items()
        for key, value in sorted(available_periods, key=lambda x: x[1]):
            if int(data['period']) < int(value):
                retention_period = key
                break

        if retention_period is None:
            retention_period = max(available_periods)[0]
            LOG.warning("Longest retention period is to short for desired"
                        " period")

        try:
            resource = self._extract_attribute(
                data['resource'], data['attribute'])
        except AttributeError:
            LOG.error("Resource: {0} does not contain attribute {1}".format(
                data['resource'], data['attribute']))
            raise

        # Granularity is optional if it is None the minimal value for InfluxDB
        # will be 1
        granularity = \
            data['granularity'] if data['granularity'] is not None else 1

        return {'db': data['db'],
                'epoch': 'ms',
                'q': self._query_format(
                    data['query'], data['aggregate'], resource, data['period'],
                    granularity, retention_period)}

    def extract_result(self, raw_results):
        """"""
        try:
            # For result structure see:
            # https://docs.openstack.org/watcher/latest/datasources/grafana.html#InfluxDB
            result = jsonutils.loads(raw_results)
            result = result['results'][0]['series'][0]
            index_aggregate = result['columns'].index(self._data['aggregate'])
            return result['values'][0][index_aggregate]
        except KeyError:
            LOG.error("Could not extract {0} for the resource: {1}".format(
                self._data['metric'], self._data['resource']))
            raise exception.NoSuchMetricForHost(
                metric=self._data['metric'], host=self._data['resource'])



# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import jsonschema

from watcher.applier.actions import base as baction
from watcher.applier.actions import migration
from watcher.common import clients
from watcher.common import exception
from watcher.common import nova_helper
from watcher.tests import base


class TestMigration(base.TestCase):

    INSTANCE_UUID = "45a37aeb-95ab-4ddb-a305-7d9f62c2f5ba"

    def setUp(self):
        super(TestMigration, self).setUp()

        self.m_osc_cls = mock.Mock()
        self.m_helper_cls = mock.Mock()
        self.m_helper = mock.Mock(spec=nova_helper.NovaHelper)
        self.m_helper_cls.return_value = self.m_helper
        self.m_osc = mock.Mock(spec=clients.OpenStackClients)
        self.m_osc_cls.return_value = self.m_osc

        m_openstack_clients = mock.patch.object(
            clients, "OpenStackClients", self.m_osc_cls)
        m_nova_helper = mock.patch.object(
            nova_helper, "NovaHelper", self.m_helper_cls)

        m_openstack_clients.start()
        m_nova_helper.start()

        self.addCleanup(m_openstack_clients.stop)
        self.addCleanup(m_nova_helper.stop)

        self.input_parameters = {
            "migration_type": "live",
            "source_node": "compute1-hostname",
            "destination_node": "compute2-hostname",
            baction.BaseAction.RESOURCE_ID: self.INSTANCE_UUID,
        }
        self.action = migration.Migrate(mock.Mock())
        self.action.input_parameters = self.input_parameters

        self.input_parameters_cold = {
            "migration_type": "cold",
            "source_node": "compute1-hostname",
            "destination_node": "compute2-hostname",
            baction.BaseAction.RESOURCE_ID: self.INSTANCE_UUID,
        }
        self.action_cold = migration.Migrate(mock.Mock())
        self.action_cold.input_parameters = self.input_parameters_cold

    def test_parameters(self):
        params = {baction.BaseAction.RESOURCE_ID:
                  self.INSTANCE_UUID,
                  self.action.MIGRATION_TYPE: 'live',
                  self.action.DESTINATION_NODE: 'compute-2',
                  self.action.SOURCE_NODE: 'compute-3'}
        self.action.input_parameters = params
        self.assertTrue(self.action.validate_parameters())

    def test_parameters_cold(self):
        params = {baction.BaseAction.RESOURCE_ID:
                  self.INSTANCE_UUID,
                  self.action.MIGRATION_TYPE: 'cold',
                  self.action.DESTINATION_NODE: 'compute-2',
                  self.action.SOURCE_NODE: 'compute-3'}
        self.action_cold.input_parameters = params
        self.assertTrue(self.action_cold.validate_parameters())

    def test_parameters_exception_empty_fields(self):
        parameters = {baction.BaseAction.RESOURCE_ID: None,
                      'migration_type': None,
                      'source_node': None,
                      'destination_node': None}
        self.action.input_parameters = parameters
        self.assertRaises(jsonschema.ValidationError,
                          self.action.validate_parameters)

    def test_parameters_exception_migration_type(self):
        parameters = {baction.BaseAction.RESOURCE_ID:
                      self.INSTANCE_UUID,
                      'migration_type': 'unknown',
                      'source_node': 'compute-2',
                      'destination_node': 'compute-3'}
        self.action.input_parameters = parameters
        self.assertRaises(jsonschema.ValidationError,
                          self.action.validate_parameters)

    def test_parameters_exception_source_node(self):
        parameters = {baction.BaseAction.RESOURCE_ID:
                      self.INSTANCE_UUID,
                      'migration_type': 'live',
                      'source_node': None,
                      'destination_node': 'compute-3'}
        self.action.input_parameters = parameters
        self.assertRaises(jsonschema.ValidationError,
                          self.action.validate_parameters)

    def test_parameters_destination_node_none(self):
        parameters = {baction.BaseAction.RESOURCE_ID:
                      self.INSTANCE_UUID,
                      'migration_type': 'live',
                      'source_node': 'compute-1',
                      'destination_node': None}
        self.action.input_parameters = parameters
        self.assertTrue(self.action.validate_parameters)

    def test_parameters_exception_resource_id(self):
        parameters = {baction.BaseAction.RESOURCE_ID: "EFEF",
                      'migration_type': 'live',
                      'source_node': 'compute-2',
                      'destination_node': 'compute-3'}
        self.action.input_parameters = parameters
        self.assertRaises(jsonschema.ValidationError,
                          self.action.validate_parameters)

    def test_migration_pre_condition(self):
        try:
            self.action.pre_condition()
        except Exception as exc:
            self.fail(exc)

    def test_migration_post_condition(self):
        try:
            self.action.post_condition()
        except Exception as exc:
            self.fail(exc)

    def test_execute_live_migration_invalid_instance(self):
        self.m_helper.find_instance.return_value = None
        exc = self.assertRaises(
            exception.InstanceNotFound, self.action.execute)
        self.m_helper.find_instance.assert_called_once_with(self.INSTANCE_UUID)
        self.assertEqual(self.INSTANCE_UUID, exc.kwargs["name"])

    def test_execute_cold_migration_invalid_instance(self):
        self.m_helper.find_instance.return_value = None
        exc = self.assertRaises(
            exception.InstanceNotFound, self.action_cold.execute)
        self.m_helper.find_instance.assert_called_once_with(self.INSTANCE_UUID)
        self.assertEqual(self.INSTANCE_UUID, exc.kwargs["name"])

    def test_execute_live_migration(self):
        self.m_helper.find_instance.return_value = self.INSTANCE_UUID

        try:
            self.action.execute()
        except Exception as exc:
            self.fail(exc)

        self.m_helper.live_migrate_instance.assert_called_once_with(
            instance_id=self.INSTANCE_UUID,
            dest_hostname="compute2-hostname")

    def test_execute_cold_migration(self):
        self.m_helper.find_instance.return_value = self.INSTANCE_UUID

        try:
            self.action_cold.execute()
        except Exception as exc:
            self.fail(exc)

        self.m_helper.watcher_non_live_migrate_instance.\
            assert_called_once_with(
                instance_id=self.INSTANCE_UUID,
                dest_hostname="compute2-hostname"
            )

    def test_revert_live_migration(self):
        self.m_helper.find_instance.return_value = self.INSTANCE_UUID

        self.action.revert()

        self.m_helper_cls.assert_called_once_with(osc=self.m_osc)
        self.m_helper.live_migrate_instance.assert_called_once_with(
            instance_id=self.INSTANCE_UUID,
            dest_hostname="compute1-hostname"
        )

    def test_revert_cold_migration(self):
        self.m_helper.find_instance.return_value = self.INSTANCE_UUID

        self.action_cold.revert()

        self.m_helper_cls.assert_called_once_with(osc=self.m_osc)
        self.m_helper.watcher_non_live_migrate_instance.\
            assert_called_once_with(
                instance_id=self.INSTANCE_UUID,
                dest_hostname="compute1-hostname"
            )

    def test_abort_live_migrate(self):
        migration = mock.MagicMock()
        migration.id = "2"
        migrations = [migration]
        self.m_helper.get_running_migration.return_value = migrations
        self.m_helper.find_instance.return_value = self.INSTANCE_UUID
        try:
            self.action.abort()
        except Exception as exc:
            self.fail(exc)

        self.m_helper.abort_live_migrate.assert_called_once_with(
            instance_id=self.INSTANCE_UUID, source="compute1-hostname",
            destination="compute2-hostname")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from watcher.common import cinder_helper
from watcher.common import exception
from watcher.decision_engine.model.collector import cinder
from watcher.tests import base
from watcher.tests import conf_fixture


class TestCinderClusterDataModelCollector(base.TestCase):

    def setUp(self):
        super(TestCinderClusterDataModelCollector, self).setUp()
        self.useFixture(conf_fixture.ConfReloadFixture())

    @mock.patch('keystoneclient.v3.client.Client', mock.Mock())
    @mock.patch.object(cinder_helper, 'CinderHelper')
    def test_cinder_cdmc_execute(self, m_cinder_helper_cls):

        m_cinder_helper = mock.Mock(name="cinder_helper")
        m_cinder_helper_cls.return_value = m_cinder_helper

        fake_storage_node = mock.Mock(
            host='host@backend',
            zone='zone',
            status='enabled',
            state='up',
            volume_type=['fake_type']
        )

        fake_storage_pool = mock.Mock(
            total_volumes=1,
            total_capacity_gb=30,
            free_capacity_gb=20,
            provisioned_capacity_gb=10,
            allocated_capacity_gb=10,
            virtual_free=20
        )
        setattr(fake_storage_pool, 'name', 'host@backend#pool')

        fake_volume = mock.Mock(
            id=1,
            size=1,
            status='in-use',
            attachments=[{"server_id": "server_id",
                          "attachment_id": "attachment_id"}],
            multiattach='false',
            snapshot_id='',
            metadata='{"key": "value"}',
            bootable='false'
        )
        setattr(fake_volume, 'name', 'name')
        setattr(fake_volume, 'os-vol-tenant-attr:tenant_id',
                '0c003652-0cb1-4210-9005-fd5b92b1faa2')
        setattr(fake_volume, 'os-vol-host-attr:host', 'host@backend#pool')

        # storage node list
        m_cinder_helper.get_storage_node_list.return_value = [
            fake_storage_node]
        m_cinder_helper.get_volume_type_by_backendname.return_value = [
            'fake_type']

        # storage pool list
        m_cinder_helper.get_storage_pool_list.return_value = [
            fake_storage_pool]
        # volume list
        m_cinder_helper.get_volume_list.return_value = [fake_volume]

        m_config = mock.Mock()
        m_osc = mock.Mock()

        cinder_cdmc = cinder.CinderClusterDataModelCollector(
            config=m_config, osc=m_osc)

        cinder_cdmc.get_audit_scope_handler([])
        model = cinder_cdmc.execute()

        storage_nodes = model.get_all_storage_nodes()
        storage_node = list(storage_nodes.values())[0]

        storage_pools = model.get_node_pools(storage_node)
        storage_pool = storage_pools[0]

        volumes = model.get_pool_volumes(storage_pool)
        volume = volumes[0]

        self.assertEqual(1, len(storage_nodes))
        self.assertEqual(1, len(storage_pools))
        self.assertEqual(1, len(volumes))

        self.assertEqual(storage_node.host, 'host@backend')
        self.assertEqual(storage_pool.name, 'host@backend#pool')
        self.assertEqual(volume.uuid, '1')

    @mock.patch('keystoneclient.v3.client.Client', mock.Mock())
    @mock.patch.object(cinder_helper, 'CinderHelper')
    def test_cinder_cdmc_total_capacity_gb_not_integer(
            self, m_cinder_helper_cls):

        m_cinder_helper = mock.Mock(name="cinder_helper")
        m_cinder_helper_cls.return_value = m_cinder_helper

        fake_storage_node = mock.Mock(
            host='host@backend',
            zone='zone',
            status='enabled',
            state='up',
            volume_type=['fake_type']
        )

        fake_storage_pool = mock.Mock(
            total_volumes=1,
            total_capacity_gb="unknown",
            free_capacity_gb=20,
            provisioned_capacity_gb=10,
            allocated_capacity_gb=10,
            virtual_free=20
        )
        setattr(fake_storage_pool, 'name', 'host@backend#pool')

        # storage node list
        m_cinder_helper.get_storage_node_list.return_value = [
            fake_storage_node]
        m_cinder_helper.get_volume_type_by_backendname.return_value = [
            'fake_type']

        # storage pool list
        m_cinder_helper.get_storage_pool_list.return_value = [
            fake_storage_pool]

        # volume list
        m_cinder_helper.get_volume_list.return_value = []

        m_config = mock.Mock()
        m_osc = mock.Mock()

        cinder_cdmc = cinder.CinderClusterDataModelCollector(
            config=m_config, osc=m_osc)

        cinder_cdmc.get_audit_scope_handler([])
        self.assertRaises(exception.InvalidPoolAttributeValue,
                          cinder_cdmc.execute)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from cinderclient import client as ciclient
from cinderclient.v3 import client as ciclient_v3
from glanceclient import client as glclient
from gnocchiclient import client as gnclient
from gnocchiclient.v1 import client as gnclient_v1
from ironicclient import client as irclient
from ironicclient.v1 import client as irclient_v1
from keystoneauth1 import adapter as ka_adapter
from keystoneauth1 import loading as ka_loading
from monascaclient import client as monclient
from monascaclient.v2_0 import client as monclient_v2
from neutronclient.neutron import client as netclient
from neutronclient.v2_0 import client as netclient_v2
from novaclient import client as nvclient

from watcher.common import clients
from watcher import conf
from watcher.tests import base

CONF = conf.CONF


class TestClients(base.TestCase):

    def _register_watcher_clients_auth_opts(self):
        _AUTH_CONF_GROUP = 'watcher_clients_auth'
        ka_loading.register_auth_conf_options(CONF, _AUTH_CONF_GROUP)
        ka_loading.register_session_conf_options(CONF, _AUTH_CONF_GROUP)
        CONF.set_override('auth_type', 'password', group=_AUTH_CONF_GROUP)

        # ka_loading.load_auth_from_conf_options(CONF, _AUTH_CONF_GROUP)
        # ka_loading.load_session_from_conf_options(CONF, _AUTH_CONF_GROUP)
        # CONF.set_override(
        #     'auth-url', 'http://server.ip:5000', group=_AUTH_CONF_GROUP)

        # If we don't clean up the _AUTH_CONF_GROUP conf options, then other
        # tests that run after this one will fail, complaining about required
        # options that _AUTH_CONF_GROUP wants.
        def cleanup_conf_from_loading():
            # oslo_config doesn't seem to allow unregistering groups through a
            # single method, so we do this instead
            CONF.reset()
            del CONF._groups[_AUTH_CONF_GROUP]

        self.addCleanup(cleanup_conf_from_loading)

        def reset_register_opts_mock(conf_obj, original_method):
            conf_obj.register_opts = original_method

        original_register_opts = CONF.register_opts
        self.addCleanup(reset_register_opts_mock,
                        CONF,
                        original_register_opts)

        expected = {'username': 'foousername',
                    'password': 'foopassword',
                    'auth_url': 'http://server.ip:5000',
                    'cafile': None,
                    'certfile': None,
                    'keyfile': None,
                    'insecure': False,
                    'user_domain_id': 'foouserdomainid',
                    'project_domain_id': 'fooprojdomainid'}

        # Because some of the conf options for auth plugins are not registered
        # until right before they are loaded, and because the method that does
        # the actual loading of the conf option values is an anonymous method
        # (see _getter method of load_from_conf_options in
        # keystoneauth1.loading.conf.py), we need to manually monkey patch
        # the register opts method so that we can override the conf values to
        # our custom values.
        def mock_register_opts(*args, **kwargs):
            ret = original_register_opts(*args, **kwargs)
            if 'group' in kwargs and kwargs['group'] == _AUTH_CONF_GROUP:
                for key, value in expected.items():
                    CONF.set_override(key, value, group=_AUTH_CONF_GROUP)
            return ret

        CONF.register_opts = mock_register_opts

    def test_get_keystone_session(self):
        self._register_watcher_clients_auth_opts()

        osc = clients.OpenStackClients()

        expected = {'username': 'foousername',
                    'password': 'foopassword',
                    'auth_url': 'http://server.ip:5000',
                    'user_domain_id': 'foouserdomainid',
                    'project_domain_id': 'fooprojdomainid'}

        sess = osc.session
        self.assertEqual(expected['auth_url'], sess.auth.auth_url)
        self.assertEqual(expected['username'], sess.auth._username)
        self.assertEqual(expected['password'], sess.auth._password)
        self.assertEqual(expected['user_domain_id'], sess.auth._user_domain_id)
        self.assertEqual(expected['project_domain_id'],
                         sess.auth._project_domain_id)

    @mock.patch.object(nvclient, 'Client')
    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_nova(self, mock_session, mock_call):
        osc = clients.OpenStackClients()
        osc._nova = None
        osc.nova()
        mock_call.assert_called_once_with(
            CONF.nova_client.api_version,
            endpoint_type=CONF.nova_client.endpoint_type,
            region_name=CONF.nova_client.region_name,
            session=mock_session)

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_nova_diff_vers(self, mock_session):
        CONF.set_override('api_version', '2.60', group='nova_client')
        osc = clients.OpenStackClients()
        osc._nova = None
        osc.nova()
        self.assertEqual('2.60', osc.nova().api_version.get_string())

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_nova_bad_min_version(self, mock_session):
        CONF.set_override('api_version', '2.47', group='nova_client')
        osc = clients.OpenStackClients()
        osc._nova = None
        ex = self.assertRaises(ValueError, osc.nova)
        self.assertIn('Invalid nova_client.api_version 2.47', str(ex))

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_nova_diff_endpoint(self, mock_session):
        CONF.set_override('endpoint_type', 'publicURL', group='nova_client')
        osc = clients.OpenStackClients()
        osc._nova = None
        osc.nova()
        self.assertEqual('publicURL', osc.nova().client.interface)

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_nova_cached(self, mock_session):
        osc = clients.OpenStackClients()
        osc._nova = None
        nova = osc.nova()
        nova_cached = osc.nova()
        self.assertEqual(nova, nova_cached)

    @mock.patch.object(glclient, 'Client')
    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_glance(self, mock_session, mock_call):
        osc = clients.OpenStackClients()
        osc._glance = None
        osc.glance()
        mock_call.assert_called_once_with(
            CONF.glance_client.api_version,
            interface=CONF.glance_client.endpoint_type,
            region_name=CONF.glance_client.region_name,
            session=mock_session)

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_glance_diff_vers(self, mock_session):
        CONF.set_override('api_version', '1', group='glance_client')
        osc = clients.OpenStackClients()
        osc._glance = None
        osc.glance()
        self.assertEqual(1.0, osc.glance().version)

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_glance_diff_endpoint(self, mock_session):
        CONF.set_override('endpoint_type',
                          'internalURL', group='glance_client')
        osc = clients.OpenStackClients()
        osc._glance = None
        osc.glance()
        self.assertEqual('internalURL', osc.glance().http_client.interface)

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_glance_cached(self, mock_session):
        osc = clients.OpenStackClients()
        osc._glance = None
        glance = osc.glance()
        glance_cached = osc.glance()
        self.assertEqual(glance, glance_cached)

    @mock.patch.object(gnclient, 'Client')
    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_gnocchi(self, mock_session, mock_call):
        osc = clients.OpenStackClients()
        osc._gnocchi = None
        osc.gnocchi()
        mock_call.assert_called_once_with(
            CONF.gnocchi_client.api_version,
            adapter_options={
                "interface": CONF.gnocchi_client.endpoint_type,
                "region_name": CONF.gnocchi_client.region_name},
            session=mock_session)

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_gnocchi_diff_vers(self, mock_session):
        # gnocchiclient currently only has one version (v1)
        CONF.set_override('api_version', '1', group='gnocchi_client')
        osc = clients.OpenStackClients()
        osc._gnocchi = None
        osc.gnocchi()
        self.assertEqual(gnclient_v1.Client, type(osc.gnocchi()))

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_gnocchi_diff_endpoint(self, mock_session):
        # gnocchiclient currently only has one version (v1)
        CONF.set_override('endpoint_type', 'publicURL', group='gnocchi_client')
        osc = clients.OpenStackClients()
        osc._gnocchi = None
        osc.gnocchi()
        self.assertEqual('publicURL', osc.gnocchi().api.interface)

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_gnocchi_cached(self, mock_session):
        osc = clients.OpenStackClients()
        osc._gnocchi = None
        gnocchi = osc.gnocchi()
        gnocchi_cached = osc.gnocchi()
        self.assertEqual(gnocchi, gnocchi_cached)

    @mock.patch.object(ciclient, 'Client')
    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_cinder(self, mock_session, mock_call):
        osc = clients.OpenStackClients()
        osc._cinder = None
        osc.cinder()
        mock_call.assert_called_once_with(
            CONF.cinder_client.api_version,
            endpoint_type=CONF.cinder_client.endpoint_type,
            region_name=CONF.cinder_client.region_name,
            session=mock_session)

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_cinder_diff_vers(self, mock_session):
        CONF.set_override('api_version', '3', group='cinder_client')
        osc = clients.OpenStackClients()
        osc._cinder = None
        osc.cinder()
        self.assertEqual(ciclient_v3.Client, type(osc.cinder()))

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_cinder_diff_endpoint(self, mock_session):
        CONF.set_override('endpoint_type',
                          'internalURL', group='cinder_client')
        osc = clients.OpenStackClients()
        osc._cinder = None
        osc.cinder()
        self.assertEqual('internalURL', osc.cinder().client.interface)

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_cinder_cached(self, mock_session):
        osc = clients.OpenStackClients()
        osc._cinder = None
        cinder = osc.cinder()
        cinder_cached = osc.cinder()
        self.assertEqual(cinder, cinder_cached)

    @mock.patch.object(netclient, 'Client')
    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_neutron(self, mock_session, mock_call):
        osc = clients.OpenStackClients()
        osc._neutron = None
        osc.neutron()
        mock_call.assert_called_once_with(
            CONF.neutron_client.api_version,
            endpoint_type=CONF.neutron_client.endpoint_type,
            region_name=CONF.neutron_client.region_name,
            session=mock_session)

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_neutron_diff_vers(self, mock_session):
        '''neutronclient currently only has one version (v2)'''
        CONF.set_override('api_version', '2.0',
                          group='neutron_client')
        osc = clients.OpenStackClients()
        osc._neutron = None
        osc.neutron()
        self.assertEqual(netclient_v2.Client,
                         type(osc.neutron()))

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_neutron_diff_endpoint(self, mock_session):
        '''neutronclient currently only has one version (v2)'''
        CONF.set_override('endpoint_type', 'internalURL',
                          group='neutron_client')
        osc = clients.OpenStackClients()
        osc._neutron = None
        osc.neutron()
        self.assertEqual('internalURL', osc.neutron().httpclient.interface)

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_neutron_cached(self, mock_session):
        osc = clients.OpenStackClients()
        osc._neutron = None
        neutron = osc.neutron()
        neutron_cached = osc.neutron()
        self.assertEqual(neutron, neutron_cached)

    @mock.patch.object(monclient, 'Client')
    @mock.patch.object(ka_loading, 'load_session_from_conf_options')
    def test_clients_monasca(self, mock_session, mock_call):
        mock_session.return_value = mock.Mock(
            get_endpoint=mock.Mock(return_value='test_endpoint'),
            get_token=mock.Mock(return_value='test_token'),)

        self._register_watcher_clients_auth_opts()

        osc = clients.OpenStackClients()
        osc._monasca = None
        osc.monasca()
        mock_call.assert_called_once_with(
            CONF.monasca_client.api_version,
            'test_endpoint',
            auth_url='http://server.ip:5000', cert_file=None, insecure=False,
            key_file=None, keystone_timeout=None, os_cacert=None,
            password='foopassword', service_type='monitoring',
            token='test_token', username='foousername')

    @mock.patch.object(ka_loading, 'load_session_from_conf_options')
    def test_clients_monasca_diff_vers(self, mock_session):
        mock_session.return_value = mock.Mock(
            get_endpoint=mock.Mock(return_value='test_endpoint'),
            get_token=mock.Mock(return_value='test_token'),)

        self._register_watcher_clients_auth_opts()

        CONF.set_override('api_version', '2_0', group='monasca_client')
        osc = clients.OpenStackClients()
        osc._monasca = None
        osc.monasca()
        self.assertEqual(monclient_v2.Client, type(osc.monasca()))

    @mock.patch.object(ka_loading, 'load_session_from_conf_options')
    def test_clients_monasca_cached(self, mock_session):
        mock_session.return_value = mock.Mock(
            get_endpoint=mock.Mock(return_value='test_endpoint'),
            get_token=mock.Mock(return_value='test_token'),)

        self._register_watcher_clients_auth_opts()

        osc = clients.OpenStackClients()
        osc._monasca = None
        monasca = osc.monasca()
        monasca_cached = osc.monasca()
        self.assertEqual(monasca, monasca_cached)

    @mock.patch.object(irclient, 'Client')
    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_ironic(self, mock_session, mock_call):
        ironic_url = 'http://localhost:6385/'
        mock_session.get_endpoint.return_value = ironic_url
        osc = clients.OpenStackClients()
        osc._ironic = None
        osc.ironic()
        mock_call.assert_called()

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_ironic_diff_vers(self, mock_session):
        ironic_url = 'http://localhost:6385/'
        mock_session.get_endpoint.return_value = ironic_url
        CONF.set_override('api_version', '1', group='ironic_client')
        osc = clients.OpenStackClients()
        osc._ironic = None
        osc.ironic()
        self.assertEqual(irclient_v1.Client, type(osc.ironic()))

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_ironic_diff_endpoint(self, mock_session):
        ironic_url = 'http://localhost:6385/'
        mock_session.get_endpoint.return_value = ironic_url
        osc = clients.OpenStackClients()
        osc._ironic = None
        osc.ironic()
        mock_session.get_endpoint.assert_called_with(
            interface='publicURL',
            region_name=None,
            service_type='baremetal')

        CONF.set_override('endpoint_type', 'internalURL',
                          group='ironic_client')
        osc._ironic = None
        osc.ironic()
        mock_session.get_endpoint.assert_called_with(
            interface='internalURL',
            region_name=None,
            service_type='baremetal')

    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_ironic_cached(self, mock_session):
        ironic_url = 'http://localhost:6385/'
        mock_session.get_endpoint.return_value = ironic_url
        osc = clients.OpenStackClients()
        osc._ironic = None
        ironic = osc.ironic()
        ironic_cached = osc.ironic()
        self.assertEqual(ironic, ironic_cached)

    @mock.patch.object(ka_adapter, 'Adapter')
    @mock.patch.object(clients.OpenStackClients, 'session')
    def test_clients_placement(self, mock_session, mock_call):
        osc = clients.OpenStackClients()
        osc.placement()
        headers = {'accept': 'application/json'}
        mock_call.assert_called_once_with(
            session=mock_session,
            service_type='placement',
            default_microversion=CONF.placement_client.api_version,
            interface=CONF.placement_client.interface,
            region_name=CONF.placement_client.region_name,
            additional_headers=headers)



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from http import HTTPStatus
from oslo_config import cfg
from oslo_serialization import jsonutils
from urllib import parse as urlparse

from watcher.tests.api import base as api_base
from watcher.tests.objects import utils as obj_utils


class TestListService(api_base.FunctionalTest):

    def _assert_service_fields(self, service):
        service_fields = ['id', 'name', 'host', 'status']
        for field in service_fields:
            self.assertIn(field, service)

    def test_one(self):
        service = obj_utils.create_test_service(self.context)
        response = self.get_json('/services')
        self.assertEqual(service.id, response['services'][0]["id"])
        self._assert_service_fields(response['services'][0])

    def test_get_one_by_id(self):
        service = obj_utils.create_test_service(self.context)
        response = self.get_json('/services/%s' % service.id)
        self.assertEqual(service.id, response["id"])
        self.assertEqual(service.name, response["name"])
        self._assert_service_fields(response)

    def test_get_one_by_name(self):
        service = obj_utils.create_test_service(self.context)
        response = self.get_json(urlparse.quote(
            '/services/%s' % service['name']))
        self.assertEqual(service.id, response['id'])
        self._assert_service_fields(response)

    def test_get_one_soft_deleted(self):
        service = obj_utils.create_test_service(self.context)
        service.soft_delete()
        response = self.get_json(
            '/services/%s' % service['id'],
            headers={'X-Show-Deleted': 'True'})
        self.assertEqual(service.id, response['id'])
        self._assert_service_fields(response)

        response = self.get_json(
            '/services/%s' % service['id'],
            expect_errors=True)
        self.assertEqual(HTTPStatus.NOT_FOUND, response.status_int)

    def test_detail(self):
        service = obj_utils.create_test_service(self.context)
        response = self.get_json('/services/detail')
        self.assertEqual(service.id, response['services'][0]["id"])
        self._assert_service_fields(response['services'][0])
        for service in response['services']:
            self.assertTrue(
                all(val is not None for key, val in service.items()
                    if key in ['id', 'name', 'host', 'status'])
            )

    def test_detail_against_single(self):
        service = obj_utils.create_test_service(self.context)
        response = self.get_json('/services/%s/detail' % service.id,
                                 expect_errors=True)
        self.assertEqual(HTTPStatus.NOT_FOUND, response.status_int)

    def test_many(self):
        service_list = []
        for idx in range(1, 4):
            service = obj_utils.create_test_service(
                self.context, id=idx, host='CONTROLLER1',
                name='SERVICE_{0}'.format(idx))
            service_list.append(service.id)
        for idx in range(1, 4):
            service = obj_utils.create_test_service(
                self.context, id=3+idx, host='CONTROLLER2',
                name='SERVICE_{0}'.format(idx))
            service_list.append(service.id)
        response = self.get_json('/services')
        self.assertEqual(6, len(response['services']))
        for service in response['services']:
            self.assertTrue(
                all(val is not None for key, val in service.items()
                    if key in ['id', 'name', 'host', 'status']))

    def test_many_without_soft_deleted(self):
        service_list = []
        for id_ in [1, 2, 3]:
            service = obj_utils.create_test_service(
                self.context, id=id_, host='CONTROLLER',
                name='SERVICE_{0}'.format(id_))
            service_list.append(service.id)
        for id_ in [4, 5]:
            service = obj_utils.create_test_service(
                self.context, id=id_, host='CONTROLLER',
                name='SERVICE_{0}'.format(id_))
            service.soft_delete()
        response = self.get_json('/services')
        self.assertEqual(3, len(response['services']))
        ids = [s['id'] for s in response['services']]
        self.assertEqual(sorted(service_list), sorted(ids))

    def test_services_collection_links(self):
        for idx in range(1, 6):
            obj_utils.create_test_service(
                self.context, id=idx,
                host='CONTROLLER',
                name='SERVICE_{0}'.format(idx))
        response = self.get_json('/services/?limit=2')
        self.assertEqual(2, len(response['services']))

    def test_services_collection_links_default_limit(self):
        for idx in range(1, 6):
            obj_utils.create_test_service(
                self.context, id=idx,
                host='CONTROLLER',
                name='SERVICE_{0}'.format(idx))
        cfg.CONF.set_override('max_limit', 3, 'api')
        response = self.get_json('/services')
        self.assertEqual(3, len(response['services']))

    def test_many_with_sort_key_name(self):
        service_list = []
        for id_ in range(1, 4):
            service = obj_utils.create_test_service(
                self.context, id=id_, host='CONTROLLER',
                name='SERVICE_{0}'.format(id_))
            service_list.append(service.name)

        response = self.get_json('/services/?sort_key=name')

        self.assertEqual(3, len(response['services']))
        names = [s['name'] for s in response['services']]
        self.assertEqual(sorted(service_list), names)

    def test_sort_key_validation(self):
        response = self.get_json(
            '/services?sort_key=%s' % 'bad_name',
            expect_errors=True)
        self.assertEqual(HTTPStatus.BAD_REQUEST, response.status_int)


class TestServicePolicyEnforcement(api_base.FunctionalTest):

    def _common_policy_check(self, rule, func, *arg, **kwarg):
        self.policy.set_rules({
            "admin_api": "(role:admin or role:administrator)",
            "default": "rule:admin_api",
            rule: "rule:default"})
        response = func(*arg, **kwarg)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(
            "Policy doesn't allow %s to be performed." % rule,
            jsonutils.loads(response.json['error_message'])['faultstring'])

    def test_policy_disallow_get_all(self):
        self._common_policy_check(
            "service:get_all", self.get_json, '/services',
            expect_errors=True)

    def test_policy_disallow_get_one(self):
        service = obj_utils.create_test_service(self.context)
        self._common_policy_check(
            "service:get", self.get_json,
            '/services/%s' % service.id,
            expect_errors=True)

    def test_policy_disallow_detail(self):
        self._common_policy_check(
            "service:detail", self.get_json,
            '/services/detail',
            expect_errors=True)


class TestServiceEnforcementWithAdminContext(TestListService,
                                             api_base.AdminRoleTest):

    def setUp(self):
        super(TestServiceEnforcementWithAdminContext, self).setUp()
        self.policy.set_rules({
            "admin_api": "(role:admin or role:administrator)",
            "default": "rule:admin_api",
            "service:detail": "rule:default",
            "service:get": "rule:default",
            "service:get_all": "rule:default"})



"""Add cron support for audit table

Revision ID: d098df6021e2
Revises: 0f6042416884
Create Date: 2017-06-08 16:21:35.746752

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'd098df6021e2'
down_revision = '0f6042416884'


def upgrade():
    op.alter_column('audits', 'interval', existing_type=sa.String(36),
                    nullable=True)
    op.add_column('audits',
                  sa.Column('next_run_time', sa.DateTime(), nullable=True))


def downgrade():
    op.alter_column('audits', 'interval', existing_type=sa.Integer(),
                    nullable=True)
    op.drop_column('audits', 'next_run_time')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from watcher.common import cinder_helper
from watcher.common import exception

from watcher.decision_engine.scope import base


class StorageScope(base.BaseScope):
    """Storage Audit Scope Handler"""

    def __init__(self, scope, config, osc=None):
        super(StorageScope, self).__init__(scope, config)
        self._osc = osc
        self.wrapper = cinder_helper.CinderHelper(osc=self._osc)

    def _collect_vtype(self, volume_types, allowed_nodes):
        service_list = self.wrapper.get_storage_node_list()

        vt_names = [volume_type['name'] for volume_type in volume_types]
        include_all_nodes = False
        if '*' in vt_names:
            if len(vt_names) == 1:
                include_all_nodes = True
            else:
                raise exception.WildcardCharacterIsUsed(
                    resource="volume_types")
        for service in service_list:
            if include_all_nodes:
                allowed_nodes.append(service.host)
                continue
            backend = service.host.split('@')[1]
            v_types = self.wrapper.get_volume_type_by_backendname(
                backend)
            for volume_type in v_types:
                if volume_type in vt_names:
                    # Note(adisky): It can generate duplicate values
                    # but it will later converted to set
                    allowed_nodes.append(service.host)

    def _collect_zones(self, availability_zones, allowed_nodes):
        service_list = self.wrapper.get_storage_node_list()
        zone_names = [zone['name'] for zone
                      in availability_zones]
        include_all_nodes = False
        if '*' in zone_names:
            if len(zone_names) == 1:
                include_all_nodes = True
            else:
                raise exception.WildcardCharacterIsUsed(
                    resource="availability zones")
        for service in service_list:
            if service.zone in zone_names or include_all_nodes:
                allowed_nodes.append(service.host)

    def exclude_resources(self, resources, **kwargs):
        pools_to_exclude = kwargs.get('pools')
        volumes_to_exclude = kwargs.get('volumes')
        projects_to_exclude = kwargs.get('projects')

        for resource in resources:
            if 'storage_pools' in resource:
                pools_to_exclude.extend(
                    [storage_pool['name'] for storage_pool
                     in resource['storage_pools']])

            elif 'volumes' in resource:
                volumes_to_exclude.extend(
                    [volume['uuid'] for volume in
                     resource['volumes']])

            elif 'projects' in resource:
                projects_to_exclude.extend(
                    [project['uuid'] for project in
                     resource['projects']])

    def exclude_pools(self, pools_to_exclude, cluster_model):
        for pool_name in pools_to_exclude:
            pool = cluster_model.get_pool_by_pool_name(pool_name)
            volumes = cluster_model.get_pool_volumes(pool)
            for volume in volumes:
                cluster_model.remove_volume(volume)
            cluster_model.remove_pool(pool)

    def exclude_volumes(self, volumes_to_exclude, cluster_model):
        for volume_uuid in volumes_to_exclude:
            volume = cluster_model.get_volume_by_uuid(volume_uuid)
            cluster_model.remove_volume(volume)

    def exclude_projects(self, projects_to_exclude, cluster_model):
        all_volumes = cluster_model.get_all_volumes()
        for volume_uuid in all_volumes:
            volume = all_volumes.get(volume_uuid)
            if volume.project_id in projects_to_exclude:
                cluster_model.remove_volume(volume)

    def remove_nodes_from_model(self, nodes_to_remove, cluster_model):
        for hostname in nodes_to_remove:
            node = cluster_model.get_node_by_name(hostname)
            pools = cluster_model.get_node_pools(node)
            for pool in pools:
                volumes = cluster_model.get_pool_volumes(pool)
                for volume in volumes:
                    cluster_model.remove_volume(volume)
                cluster_model.remove_pool(pool)
            cluster_model.remove_node(node)

    def get_scoped_model(self, cluster_model):
        """Leave only nodes, pools and volumes proposed in the audit scope"""
        if not cluster_model:
            return None

        allowed_nodes = []
        nodes_to_remove = set()
        volumes_to_exclude = []
        projects_to_exclude = []
        pools_to_exclude = []

        model_hosts = list(cluster_model.get_all_storage_nodes().keys())

        storage_scope = []

        for scope in self.scope:
            storage_scope = scope.get('storage')
            if storage_scope:
                break

        if not storage_scope:
            return cluster_model

        for rule in storage_scope:
            if 'volume_types' in rule:
                self._collect_vtype(rule['volume_types'],
                                    allowed_nodes, cluster_model)
            elif 'availability_zones' in rule:
                self._collect_zones(rule['availability_zones'],
                                    allowed_nodes)
            elif 'exclude' in rule:
                self.exclude_resources(
                    rule['exclude'], pools=pools_to_exclude,
                    volumes=volumes_to_exclude,
                    projects=projects_to_exclude)

        if allowed_nodes:
            nodes_to_remove = set(model_hosts) - set(allowed_nodes)

        self.remove_nodes_from_model(nodes_to_remove, cluster_model)
        self.exclude_pools(pools_to_exclude, cluster_model)
        self.exclude_volumes(volumes_to_exclude, cluster_model)
        self.exclude_projects(projects_to_exclude, cluster_model)

        return cluster_model



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from watcher.decision_engine.strategy.strategies import base as base_strategy


class FakeStrategy(base_strategy.BaseStrategy):

    NAME = NotImplemented
    DISPLAY_NAME = NotImplemented
    GOAL_NAME = NotImplemented

    @classmethod
    def get_name(cls):
        return cls.NAME

    @classmethod
    def get_display_name(cls):
        return cls.DISPLAY_NAME

    @classmethod
    def get_translatable_display_name(cls):
        return cls.DISPLAY_NAME

    @classmethod
    def get_goal_name(cls):
        return cls.GOAL_NAME

    @classmethod
    def get_config_opts(cls):
        return []

    def pre_execute(self):
        pass

    def do_execute(self):
        pass

    def post_execute(self):
        pass


class FakeDummy1Strategy1(FakeStrategy):
    GOAL_NAME = "dummy_1"
    NAME = "strategy_1"
    DISPLAY_NAME = "Strategy 1"

    @classmethod
    def get_config_opts(cls):
        return [
            cfg.StrOpt('test_opt', help="Option used for testing."),
        ]


class FakeDummy1Strategy2(FakeStrategy):
    GOAL_NAME = "dummy_1"
    NAME = "strategy_2"
    DISPLAY_NAME = "Strategy 2"


class FakeDummy2Strategy3(FakeStrategy):
    GOAL_NAME = "dummy_2"
    NAME = "strategy_3"
    DISPLAY_NAME = "Strategy 3"


class FakeDummy2Strategy4(FakeStrategy):
    GOAL_NAME = "dummy_2"
    NAME = "strategy_4"
    DISPLAY_NAME = "Strategy 4"



# -*- encoding: utf-8 -*-
# Copyright (c) 2017 ZTE
#
# Authors: Canwei Li <li.canwei2@zte.com.cn>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from unittest import mock


from watcher.common import cinder_helper
from watcher.common import clients
from watcher.common import utils
from watcher.decision_engine.strategy import strategies
from watcher.tests.decision_engine.model import faker_cluster_state
from watcher.tests.decision_engine.strategy.strategies.test_base \
    import TestBaseStrategy


class TestStorageCapacityBalance(TestBaseStrategy):

    def setUp(self):
        super(TestStorageCapacityBalance, self).setUp()

        def test_fake_pool(name, free, total, allocated):
            fake_pool = mock.MagicMock()
            fake_pool.name = name
            fake_pool.pool_name = name.split('#')[1]
            fake_pool.volume_backend_name = name.split('#')[1]
            fake_pool.free_capacity_gb = free
            fake_pool.total_capacity_gb = total
            fake_pool.allocated_capacity_gb = allocated
            fake_pool.max_over_subscription_ratio = 1.0

            return fake_pool

        self.fake_pool1 = test_fake_pool('host1@IPSAN-1#pool1',
                                         '60', '100', '90')

        self.fake_pool2 = test_fake_pool('host1@IPSAN-1#pool2',
                                         '20', '100', '80')

        self.fake_pool3 = test_fake_pool('host1@IPSAN-1#local_vstorage',
                                         '20', '100', '80')
        self.fake_pools = [self.fake_pool1, self.fake_pool2,
                           self.fake_pool3]

        def test_fake_vol(id, name, size, status, bootable,
                          migration_status=None,
                          volume_type=None):
            fake_vol = mock.MagicMock()
            fake_vol.id = id
            fake_vol.name = name
            fake_vol.size = size
            fake_vol.status = status
            fake_vol.bootable = bootable
            fake_vol.migration_status = migration_status
            fake_vol.volume_type = volume_type
            setattr(fake_vol, 'os-vol-host-attr:host', 'host1@IPSAN-1#pool2')

            return fake_vol

        self.fake_vol1 = test_fake_vol('922d4762-0bc5-4b30-9cb9-48ab644dd861',
                                       'test_volume1', 4,
                                       'available', 'true', 'success',
                                       volume_type='type2')
        self.fake_vol2 = test_fake_vol('922d4762-0bc5-4b30-9cb9-48ab644dd862',
                                       'test_volume2', 10,
                                       'in-use', 'false')
        self.fake_vol3 = test_fake_vol('922d4762-0bc5-4b30-9cb9-48ab644dd863',
                                       'test_volume3', 4,
                                       'in-use', 'true', volume_type='type2')
        self.fake_vol4 = test_fake_vol('922d4762-0bc5-4b30-9cb9-48ab644dd864',
                                       'test_volume4', 10,
                                       'error', 'true')
        self.fake_vol5 = test_fake_vol('922d4762-0bc5-4b30-9cb9-48ab644dd865',
                                       'test_volume5', 15,
                                       'in-use', 'true')

        self.fake_volumes = [self.fake_vol1,
                             self.fake_vol2,
                             self.fake_vol3,
                             self.fake_vol4,
                             self.fake_vol5]

        def test_fake_snap(vol_id):
            fake_snap = mock.MagicMock()
            fake_snap.volume_id = vol_id

            return fake_snap

        self.fake_snap = [test_fake_snap(
            '922d4762-0bc5-4b30-9cb9-48ab644dd865')]

        def test_fake_volume_type(type_name, extra_specs):
            fake_type = mock.MagicMock()
            fake_type.name = type_name
            fake_type.extra_specs = extra_specs

            return fake_type

        self.fake_types = [test_fake_volume_type(
            'type1', {'volume_backend_name': 'pool1'}),
            test_fake_volume_type(
            'type2', {'volume_backend_name': 'pool2'})
            ]

        self.fake_c_cluster = faker_cluster_state.FakerStorageModelCollector()

        osc = clients.OpenStackClients()

        p_cinder = mock.patch.object(osc, 'cinder')
        p_cinder.start()
        self.addCleanup(p_cinder.stop)
        self.m_cinder = cinder_helper.CinderHelper(osc=osc)

        self.m_cinder.get_storage_pool_list = mock.Mock(
            return_value=self.fake_pools)
        self.m_cinder.get_volume_list = mock.Mock(
            return_value=self.fake_volumes)
        self.m_cinder.get_volume_snapshots_list = mock.Mock(
            return_value=self.fake_snap)
        self.m_cinder.get_volume_type_list = mock.Mock(
            return_value=self.fake_types)

        model = self.fake_c_cluster.generate_scenario_1()
        self.m_c_model.return_value = model

        self.strategy = strategies.StorageCapacityBalance(
            config=mock.Mock(), osc=osc)
        self.strategy._cinder = self.m_cinder
        self.strategy.input_parameters = utils.Struct()
        self.strategy.input_parameters.update(
            {'volume_threshold': 80.0})
        self.strategy.volume_threshold = 80.0

    def test_get_pools(self):
        self.strategy.config.ex_pools = "local_vstorage"
        pools = self.strategy.get_pools(self.m_cinder)
        self.assertEqual(len(pools), 2)

    def test_get_volumes(self):
        volumes = self.strategy.get_volumes(self.m_cinder)
        self.assertEqual(len(volumes), 3)

    def test_group_pools(self):
        self.strategy.config.ex_pools = "local_vstorage"
        pools = self.strategy.get_pools(self.m_cinder)
        over_pools, under_pools = self.strategy.group_pools(pools, 0.50)
        self.assertEqual(len(under_pools), 1)
        self.assertEqual(len(over_pools), 1)

        over_pools, under_pools = self.strategy.group_pools(pools, 0.85)
        self.assertEqual(len(under_pools), 2)
        self.assertEqual(len(over_pools), 0)

        over_pools, under_pools = self.strategy.group_pools(pools, 0.30)
        self.assertEqual(len(under_pools), 0)
        self.assertEqual(len(over_pools), 2)

    def test_get_volume_type_by_name(self):
        vol_type = self.strategy.get_volume_type_by_name(
            self.m_cinder, 'pool1')
        self.assertEqual(len(vol_type), 1)

        vol_type = self.strategy.get_volume_type_by_name(
            self.m_cinder, 'ks3200')
        self.assertEqual(len(vol_type), 0)

    def test_check_pool_type(self):
        pool_type = self.strategy.check_pool_type(
            self.fake_vol3, self.fake_pool1)
        self.assertIsNotNone(pool_type)

        pool_type = self.strategy.check_pool_type(
            self.fake_vol3, self.fake_pool2)
        self.assertIsNone(pool_type)

    def test_migrate_fit(self):
        self.strategy.config.ex_pools = "local_vstorage"
        pools = self.strategy.get_pools(self.m_cinder)
        self.strategy.source_pools, self.strategy.dest_pools = (
            self.strategy.group_pools(pools, 0.60))
        target_pool = self.strategy.migrate_fit(self.fake_vol2, 0.60)
        self.assertIsNotNone(target_pool)

        target_pool = self.strategy.migrate_fit(self.fake_vol3, 0.50)
        self.assertIsNone(target_pool)

        target_pool = self.strategy.migrate_fit(self.fake_vol5, 0.60)
        self.assertIsNone(target_pool)

    def test_retype_fit(self):
        self.strategy.config.ex_pools = "local_vstorage"
        pools = self.strategy.get_pools(self.m_cinder)
        self.strategy.source_pools, self.strategy.dest_pools = (
            self.strategy.group_pools(pools, 0.50))
        target_pool = self.strategy.retype_fit(self.fake_vol1, 0.50)
        self.assertIsNotNone(target_pool)

        target_pool = self.strategy.retype_fit(self.fake_vol2, 0.50)
        self.assertIsNone(target_pool)

        target_pool = self.strategy.retype_fit(self.fake_vol3, 0.50)
        self.assertIsNotNone(target_pool)

        target_pool = self.strategy.retype_fit(self.fake_vol5, 0.60)
        self.assertIsNone(target_pool)

    def test_execute(self):
        self.strategy.input_parameters.update(
            {'volume_threshold': 45.0})
        self.strategy.config.ex_pools = "local_vstorage"
        solution = self.strategy.execute()
        self.assertEqual(len(solution.actions), 1)

        setattr(self.fake_pool1, 'free_capacity_gb', '60')
        self.strategy.input_parameters.update(
            {'volume_threshold': 50.0})
        solution = self.strategy.execute()
        self.assertEqual(len(solution.actions), 2)

        setattr(self.fake_pool1, 'free_capacity_gb', '60')
        self.strategy.input_parameters.update(
            {'volume_threshold': 60.0})

        solution = self.strategy.execute()
        self.assertEqual(len(solution.actions), 3)



"""add_start_end_time

Revision ID: 4b16194c56bc
Revises: 52804f2498c4
Create Date: 2018-03-23 00:36:29.031259

"""

# revision identifiers, used by Alembic.
revision = '4b16194c56bc'
down_revision = '52804f2498c4'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('audits', sa.Column('start_time', sa.DateTime(), nullable=True))
    op.add_column('audits', sa.Column('end_time', sa.DateTime(), nullable=True))


def downgrade():
    op.drop_column('audits', 'start_time')
    op.drop_column('audits', 'end_time')



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 Intel Corp
#
# Authors: Prudhvi Rao Shedimbi <prudhvi.rao.shedimbi@intel.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

watcher_applier = cfg.OptGroup(name='watcher_applier',
                               title='Options for the Applier messaging '
                               'core')

APPLIER_MANAGER_OPTS = [
    cfg.IntOpt('workers',
               default=1,
               min=1,
               required=True,
               help='Number of workers for applier, default value is 1.'),
    cfg.StrOpt('conductor_topic',
               default='watcher.applier.control',
               help='The topic name used for '
                    'control events, this topic '
                    'used for rpc call '),
    cfg.StrOpt('publisher_id',
               default='watcher.applier.api',
               help='The identifier used by watcher '
                    'module on the message broker'),
    cfg.StrOpt('workflow_engine',
               default='taskflow',
               required=True,
               help='Select the engine to use to execute the workflow'),
]

APPLIER_OPTS = [
    cfg.BoolOpt('rollback_when_actionplan_failed',
                default=False,
                help='If set True, the failed actionplan will rollback '
                     'when executing. Defaule value is False.'),
]


def register_opts(conf):
    conf.register_group(watcher_applier)
    conf.register_opts(APPLIER_MANAGER_OPTS, group=watcher_applier)
    conf.register_opts(APPLIER_OPTS, group=watcher_applier)


def list_opts():
    return [(watcher_applier, APPLIER_MANAGER_OPTS),
            (watcher_applier, APPLIER_OPTS)]



# -*- encoding: utf-8 -*-
# Copyright (c) 2017 Servionica
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime
import itertools
from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils

from watcher.common import context as watcher_context
from watcher.common import scheduling
from watcher import notifications

from watcher import objects

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class APISchedulingService(scheduling.BackgroundSchedulerService):

    def __init__(self, gconfig={}, **options):
        self.services_status = {}
        super(APISchedulingService, self).__init__(gconfig, **options)

    def get_services_status(self, context):
        services = objects.service.Service.list(context)
        active_s = objects.service.ServiceStatus.ACTIVE
        failed_s = objects.service.ServiceStatus.FAILED
        for service in services:
            result = self.get_service_status(context, service.id)
            if service.id not in self.services_status:
                self.services_status[service.id] = result
                continue
            if self.services_status[service.id] != result:
                self.services_status[service.id] = result
                notifications.service.send_service_update(context, service,
                                                          state=result)
                if (result == failed_s) and (
                        service.name == 'watcher-decision-engine'):
                    audit_filters = {
                        'audit_type': objects.audit.AuditType.CONTINUOUS.value,
                        'state': objects.audit.State.ONGOING,
                        'hostname': service.host
                    }
                    ongoing_audits = objects.Audit.list(
                        context,
                        filters=audit_filters,
                        eager=True)
                    alive_services = [
                        s.host for s in services
                        if (self.services_status[s.id] == active_s and
                            s.name == 'watcher-decision-engine')]

                    round_robin = itertools.cycle(alive_services)
                    for audit in ongoing_audits:
                        audit.hostname = round_robin.__next__()
                        audit.save()
                        LOG.info('Audit %(audit)s has been migrated to '
                                 '%(host)s since %(failed_host)s is in'
                                 ' %(state)s',
                                 {'audit': audit.uuid,
                                  'host': audit.hostname,
                                  'failed_host': service.host,
                                  'state': failed_s})

    def get_service_status(self, context, service_id):
        service = objects.Service.get(context, service_id)
        last_heartbeat = (service.last_seen_up or service.updated_at or
                          service.created_at)
        if isinstance(last_heartbeat, str):
            # NOTE(russellb) If this service came in over rpc via
            # conductor, then the timestamp will be a string and needs to be
            # converted back to a datetime.
            last_heartbeat = timeutils.parse_strtime(last_heartbeat)
        else:
            # Objects have proper UTC timezones, but the timeutils comparison
            # below does not (and will fail)
            last_heartbeat = last_heartbeat.replace(tzinfo=None)
        elapsed = timeutils.delta_seconds(last_heartbeat, timeutils.utcnow())
        is_up = abs(elapsed) <= CONF.service_down_time
        if not is_up:
            LOG.warning('Seems service %(name)s on host %(host)s is down. '
                        'Last heartbeat was %(lhb)s. Elapsed time is %(el)s',
                        {'name': service.name,
                         'host': service.host,
                         'lhb': str(last_heartbeat), 'el': str(elapsed)})
            return objects.service.ServiceStatus.FAILED

        return objects.service.ServiceStatus.ACTIVE

    def start(self):
        """Start service."""
        context = watcher_context.make_context(is_admin=True)
        self.add_job(self.get_services_status, name='service_status',
                     trigger='interval', jobstore='default', args=[context],
                     next_run_time=datetime.datetime.now(),
                     seconds=CONF.periodic_interval)
        super(APISchedulingService, self).start()

    def stop(self):
        """Stop service."""
        self.shutdown()

    def wait(self):
        """Wait for service to complete."""

    def reset(self):
        """Reset service.

        Called in case service running in daemon mode receives SIGHUP.
        """



# -*- encoding: utf-8 -*-
# Copyright (c) 2019 European Organization for Nuclear Research (CERN)
#
# Authors: Corne Lukken <info@dantalion.nl>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import futurist
from futurist import waiters

from oslo_config import cfg
from oslo_log import log
from oslo_service import service

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class DecisionEngineThreadPool(object, metaclass=service.Singleton):
    """Singleton threadpool to submit general tasks to"""

    def __init__(self):
        self.amount_workers = CONF.watcher_decision_engine.max_general_workers
        self._threadpool = futurist.GreenThreadPoolExecutor(
            max_workers=self.amount_workers)

    def submit(self, fn, *args, **kwargs):
        """Will submit the job to the underlying threadpool

        :param fn: function to execute in another thread
        :param args: arguments for the function
        :param kwargs: amount of arguments for the function
        :return: future to monitor progress of execution
        :rtype: :py:class"`futurist.GreenFuture`
        """

        return self._threadpool.submit(fn, *args, **kwargs)

    @staticmethod
    def do_while_futures(futures, fn, *args, **kwargs):
        """Do while to execute a function upon completion from a collection

        Will execute the specified function with its arguments when one of the
        futures from the passed collection finishes. Additionally, the future
        is passed as first argument to the function. Does not modify the passed
        collection of futures.

        :param futures: list, set or dictionary of futures
        :type  futures: list :py:class:`futurist.GreenFuture`
        :param fn:  function to execute upon the future finishing exection
        :param args: arguments for the function
        :param kwargs: amount of arguments for the function
        """

        # shallow copy the collection to not modify it outside of this method.
        # shallow copy must be used because the type of collection needs to be
        # determined at runtime (can be both list, set and dict).
        futures = copy.copy(futures)

        DecisionEngineThreadPool.do_while_futures_modify(
            futures, fn, *args, **kwargs)

    @staticmethod
    def do_while_futures_modify(futures, fn, *args, **kwargs):
        """Do while to execute a function upon completion from a collection

        Will execute the specified function with its arguments when one of the
        futures from the passed collection finishes. Additionally, the future
        is passed as first argument to the function. Modifies the collection
        by removing completed elements,

        :param futures: list, set or dictionary of futures
        :type  futures: list :py:class:`futurist.GreenFuture`
        :param fn:  function to execute upon the future finishing exection
        :param args: arguments for the function
        :param kwargs: amount of arguments for the function
        """

        waits = waiters.wait_for_any(futures)
        while len(waits[0]) > 0 or len(waits[1]) > 0:
            for future in waiters.wait_for_any(futures)[0]:
                fn(future, *args, **kwargs)
                futures.remove(future)
            waits = waiters.wait_for_any(futures)



# -*- encoding: utf-8 -*-
# Copyright (c) 2019 ZTE Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An Interface for users and admin to List Data Model.
"""

import pecan
from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from watcher.api.controllers.v1 import types
from watcher.api.controllers.v1 import utils
from watcher.common import exception
from watcher.common import policy
from watcher.decision_engine import rpcapi


class DataModelController(rest.RestController):
    """REST controller for data model"""
    def __init__(self):
        super(DataModelController, self).__init__()

    from_data_model = False
    """A flag to indicate if the requests to this controller are coming
    from the top-level resource DataModel."""

    @wsme_pecan.wsexpose(wtypes.text, wtypes.text, types.uuid)
    def get_all(self, data_model_type='compute', audit_uuid=None):
        """Retrieve information about the given data model.

        :param data_model_type: The type of data model user wants to list.
                                Supported values: compute.
                                Future support values: storage, baremetal.
                                The default value is compute.
        :param audit_uuid: The UUID of the audit,  used to filter data model
                           by the scope in audit.
        """
        if not utils.allow_list_datamodel():
            raise exception.NotAcceptable
        if self.from_data_model:
            raise exception.OperationNotPermitted
        allowed_data_model_type = [
            'compute',
            ]
        if data_model_type not in allowed_data_model_type:
            raise exception.DataModelTypeNotFound(
                data_model_type=data_model_type)
        context = pecan.request.context
        de_client = rpcapi.DecisionEngineAPI()
        policy.enforce(context, 'data_model:get_all',
                       action='data_model:get_all')
        rpc_all_data_model = de_client.get_data_model_info(
            context,
            data_model_type,
            audit_uuid)
        return rpc_all_data_model



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 Intel
#
# Authors: Tomasz Kaczynski <tomasz.kaczynski@intel.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log
from oslo_serialization import jsonutils
from oslo_utils import units

from watcher._i18n import _
from watcher.decision_engine.scoring import base

LOG = log.getLogger(__name__)


class DummyScorer(base.ScoringEngine):
    """Sample Scoring Engine implementing simplified workload classification.

    Typically a scoring engine would be implemented using machine learning
    techniques. For example, for workload classification problem the solution
    could consist of the following steps:

    1. Define a problem to solve: we want to detect the workload on the
       machine based on the collected metrics like power consumption,
       temperature, CPU load, memory usage, disk usage, network usage, etc.
    2. The workloads could be predefined, e.g. IDLE, CPU-INTENSIVE,
       MEMORY-INTENSIVE, IO-BOUND, ...
       Or we could let the ML algorithm to find the workloads based on the
       learning data provided. The decision here leads to learning algorithm
       used (supervised vs. non-supervised learning).
    3. Collect metrics from sample servers (learning data).
    4. Define the analytical model, pick ML framework and algorithm.
    5. Apply learning data to the data model. Once taught, the data model
       becomes a scoring engine and can start doing predictions or
       classifications.
    6. Wrap up the scoring engine with the class like this one, so it has a
       standard interface and can be used inside Watcher.

    This class is a greatly very simplified version of the above model. The
    goal is to provide an example how such class could be implemented and used
    in Watcher, without adding additional dependencies like machine learning
    frameworks (which can be quite heavy) or over-complicating it's internal
    implementation, which can distract from looking at the overall picture.

    That said, this class implements a workload classification "manually"
    (in plain python code) and is not intended to be used in production.
    """

    # Constants defining column indices for the input data
    PROCESSOR_TIME_PERC = 0
    MEM_TOTAL_BYTES = 1
    MEM_AVAIL_BYTES = 2
    MEM_PAGE_READS_PER_SEC = 3
    MEM_PAGE_WRITES_PER_SEC = 4
    DISK_READ_BYTES_PER_SEC = 5
    DISK_WRITE_BYTES_PER_SEC = 6
    NET_BYTES_RECEIVED_PER_SEC = 7
    NET_BYTES_SENT_PER_SEC = 8

    # Types of workload
    WORKLOAD_IDLE = 0
    WORKLOAD_CPU = 1
    WORKLOAD_MEM = 2
    WORKLOAD_DISK = 3

    def get_name(self):
        return 'dummy_scorer'

    def get_description(self):
        return 'Dummy workload classifier'

    def get_metainfo(self):
        """Metadata about input/output format of this scoring engine.

        This information is used in strategy using this scoring engine to
        prepare the input information and to understand the results.
        """

        return """{
            "feature_columns": [
                "proc-processor-time-%",
                "mem-total-bytes",
                "mem-avail-bytes",
                "mem-page-reads/sec",
                "mem-page-writes/sec",
                "disk-read-bytes/sec",
                "disk-write-bytes/sec",
                "net-bytes-received/sec",
                "net-bytes-sent/sec"],
            "result_columns": [
                "workload",
                "idle-probability",
                "cpu-probability",
                "memory-probability",
                "disk-probability"],
            "workloads": [
                "idle",
                "cpu-intensive",
                "memory-intensive",
                "disk-intensive"]
            }"""

    def calculate_score(self, features):
        """Arbitrary algorithm calculating the score.

        It demonstrates how to parse the input data (features) and serialize
        the results. It detects the workload type based on the metrics and
        also returns the probabilities of each workload detection (again,
        the arbitrary values are returned, just for demonstration how the
        "real" machine learning algorithm could work. For example, the
        Gradient Boosting Machine from H2O framework is using exactly the
        same format:
        http://www.h2o.ai/verticals/algos/gbm/
        """

        LOG.debug('Calculating score, features: %s', features)

        # By default IDLE workload will be returned
        workload = self.WORKLOAD_IDLE
        idle_prob = 0.0
        cpu_prob = 0.0
        mem_prob = 0.0
        disk_prob = 0.0

        # Basic input validation
        try:
            flist = jsonutils.loads(features)
        except Exception as e:
            raise ValueError(_('Unable to parse features: ') % e)
        if type(flist) is not list:
            raise ValueError(_('JSON list expected in feature argument'))
        if len(flist) != 9:
            raise ValueError(_('Invalid number of features, expected 9'))

        # Simple logic for workload classification
        if flist[self.PROCESSOR_TIME_PERC] >= 80:
            workload = self.WORKLOAD_CPU
            cpu_prob = 100.0
        elif flist[self.MEM_PAGE_READS_PER_SEC] >= 1000 \
                and flist[self.MEM_PAGE_WRITES_PER_SEC] >= 1000:
            workload = self.WORKLOAD_MEM
            mem_prob = 100.0
        elif flist[self.DISK_READ_BYTES_PER_SEC] >= 50*units.Mi \
                and flist[self.DISK_WRITE_BYTES_PER_SEC] >= 50*units.Mi:
            workload = self.WORKLOAD_DISK
            disk_prob = 100.0
        else:
            idle_prob = 100.0
            if flist[self.PROCESSOR_TIME_PERC] >= 40:
                cpu_prob = 50.0
            if flist[self.MEM_PAGE_READS_PER_SEC] >= 500 \
                    or flist[self.MEM_PAGE_WRITES_PER_SEC] >= 500:
                mem_prob = 50.0

        return jsonutils.dumps(
            [workload, idle_prob, cpu_prob, mem_prob, disk_prob])



# -*- encoding: utf-8 -*-
# Copyright (c) 2017 ZTE Corporation
#
# Authors:Yumeng Bao <bao.yumeng@zte.com.cn>

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from oslo_log import log

from watcher.common import clients
from watcher.common import exception
from watcher.common import utils

LOG = log.getLogger(__name__)


class IronicHelper(object):

    def __init__(self, osc=None):
        """:param osc: an OpenStackClients instance"""
        self.osc = osc if osc else clients.OpenStackClients()
        self.ironic = self.osc.ironic()

    def get_ironic_node_list(self):
        return self.ironic.node.list()

    def get_ironic_node_by_uuid(self, node_uuid):
        """Get ironic node by node UUID"""
        try:
            node = self.ironic.node.get(utils.Struct(uuid=node_uuid))
            if not node:
                raise exception.IronicNodeNotFound(uuid=node_uuid)
        except Exception as exc:
            LOG.exception(exc)
            raise exception.IronicNodeNotFound(uuid=node_uuid)
        # We need to pass an object with an 'uuid' attribute to make it work
        return node



# -*- encoding: utf-8 -*-
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from apscheduler.schedulers import background
import datetime
import freezegun
from unittest import mock

from watcher.api import scheduling
from watcher.notifications import service
from watcher import objects
from watcher.tests import base
from watcher.tests.db import base as db_base
from watcher.tests.db import utils


class TestSchedulingService(base.TestCase):

    @mock.patch.object(background.BackgroundScheduler, 'start')
    def test_start_scheduling_service(self, m_start):
        scheduler = scheduling.APISchedulingService()
        scheduler.start()
        m_start.assert_called_once_with(scheduler)
        jobs = scheduler.get_jobs()
        self.assertEqual(1, len(jobs))


class TestSchedulingServiceFunctions(db_base.DbTestCase):

    def setUp(self):
        super(TestSchedulingServiceFunctions, self).setUp()
        fake_service = utils.get_test_service(
            created_at=datetime.datetime.utcnow())
        self.fake_service = objects.Service(**fake_service)

    @mock.patch.object(scheduling.APISchedulingService, 'get_service_status')
    @mock.patch.object(objects.Service, 'list')
    @mock.patch.object(service, 'send_service_update')
    def test_get_services_status_without_services_in_list(
            self, mock_service_update, mock_get_list, mock_service_status):
        scheduler = scheduling.APISchedulingService()
        mock_get_list.return_value = [self.fake_service]
        mock_service_status.return_value = 'ACTIVE'
        scheduler.get_services_status(mock.ANY)
        mock_service_status.assert_called_once_with(mock.ANY,
                                                    self.fake_service.id)

        mock_service_update.assert_not_called()

    @mock.patch.object(scheduling.APISchedulingService, 'get_service_status')
    @mock.patch.object(objects.Service, 'list')
    @mock.patch.object(service, 'send_service_update')
    def test_get_services_status_with_services_in_list_same_status(
            self, mock_service_update, mock_get_list, mock_service_status):
        scheduler = scheduling.APISchedulingService()
        mock_get_list.return_value = [self.fake_service]
        scheduler.services_status = {1: 'ACTIVE'}
        mock_service_status.return_value = 'ACTIVE'
        scheduler.get_services_status(mock.ANY)
        mock_service_status.assert_called_once_with(mock.ANY,
                                                    self.fake_service.id)

        mock_service_update.assert_not_called()

    @mock.patch.object(scheduling.APISchedulingService, 'get_service_status')
    @mock.patch.object(objects.Service, 'list')
    @mock.patch.object(service, 'send_service_update')
    def test_get_services_status_with_services_in_list_diff_status(
            self, mock_service_update, mock_get_list, mock_service_status):
        scheduler = scheduling.APISchedulingService()
        mock_get_list.return_value = [self.fake_service]
        scheduler.services_status = {1: 'FAILED'}
        mock_service_status.return_value = 'ACTIVE'
        scheduler.get_services_status(mock.ANY)
        mock_service_status.assert_called_once_with(mock.ANY,
                                                    self.fake_service.id)

        mock_service_update.assert_called_once_with(mock.ANY,
                                                    self.fake_service,
                                                    state='ACTIVE')

    @mock.patch.object(objects.Service, 'get')
    def test_get_service_status_failed_service(
            self, mock_get):
        scheduler = scheduling.APISchedulingService()
        mock_get.return_value = self.fake_service
        service_status = scheduler.get_service_status(mock.ANY,
                                                      self.fake_service.id)
        mock_get.assert_called_once_with(mock.ANY,
                                         self.fake_service.id)
        self.assertEqual('FAILED', service_status)

    @freezegun.freeze_time('2016-09-22T08:32:26.219414')
    @mock.patch.object(objects.Service, 'get')
    def test_get_service_status_failed_active(
            self, mock_get):
        scheduler = scheduling.APISchedulingService()
        mock_get.return_value = self.fake_service
        service_status = scheduler.get_service_status(mock.ANY,
                                                      self.fake_service.id)
        mock_get.assert_called_once_with(mock.ANY,
                                         self.fake_service.id)
        self.assertEqual('ACTIVE', service_status)



# -*- encoding: utf-8 -*-
# Copyright (c) 2017  ZTE Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from oslo_config import cfg
from oslo_log import log

from watcher._i18n import _
from watcher.common import cinder_helper
from watcher.decision_engine.strategy.strategies import base

LOG = log.getLogger(__name__)


class StorageCapacityBalance(base.WorkloadStabilizationBaseStrategy):
    """Storage capacity balance using cinder volume migration

    *Description*

    This strategy migrates volumes based on the workload of the
    cinder pools.
    It makes decision to migrate a volume whenever a pool's used
    utilization % is higher than the specified threshold. The volume
    to be moved should make the pool close to average workload of all
    cinder pools.

    *Requirements*

    * You must have at least 2 cinder volume pools to run
      this strategy.

    *Limitations*

    * Volume migration depends on the storage device.
      It may take a long time.

    *Spec URL*

    http://specs.openstack.org/openstack/watcher-specs/specs/queens/implemented/storage-capacity-balance.html
    """

    def __init__(self, config, osc=None):
        """VolumeMigrate using cinder volume migration

        :param config: A mapping containing the configuration of this strategy
        :type config: :py:class:`~.Struct` instance
        :param osc: :py:class:`~.OpenStackClients` instance
        """
        super(StorageCapacityBalance, self).__init__(config, osc)
        self._cinder = None
        self.volume_threshold = 80.0
        self.pool_type_cache = dict()
        self.source_pools = []
        self.dest_pools = []

    @property
    def cinder(self):
        if not self._cinder:
            self._cinder = cinder_helper.CinderHelper(osc=self.osc)
        return self._cinder

    @classmethod
    def get_name(cls):
        return "storage_capacity_balance"

    @classmethod
    def get_display_name(cls):
        return _("Storage Capacity Balance Strategy")

    @classmethod
    def get_translatable_display_name(cls):
        return "Storage Capacity Balance Strategy"

    @classmethod
    def get_schema(cls):
        # Mandatory default setting for each element
        return {
            "properties": {
                "volume_threshold": {
                    "description": "volume threshold for capacity balance",
                    "type": "number",
                    "default": 80.0
                },
            },
        }

    @classmethod
    def get_config_opts(cls):
        return super(StorageCapacityBalance, cls).get_config_opts() + [
            cfg.ListOpt(
                "ex_pools",
                help="exclude pools",
                default=['local_vstorage']),
        ]

    def get_pools(self, cinder):
        """Get all volume pools excepting ex_pools.

        :param cinder: cinder client
        :return: volume pools
        """
        ex_pools = self.config.ex_pools
        pools = cinder.get_storage_pool_list()
        filtered_pools = [p for p in pools
                          if p.pool_name not in ex_pools]
        return filtered_pools

    def get_volumes(self, cinder):
        """Get all volumes with status in available or in-use and no snapshot.

        :param cinder: cinder client
        :return: all volumes
        """
        all_volumes = cinder.get_volume_list()
        valid_status = ['in-use', 'available']

        volume_snapshots = cinder.get_volume_snapshots_list()
        snapshot_volume_ids = []
        for snapshot in volume_snapshots:
            snapshot_volume_ids.append(snapshot.volume_id)

        nosnap_volumes = list(filter(lambda v: v.id not in snapshot_volume_ids,
                              all_volumes))
        LOG.info("volumes in snap: %s", snapshot_volume_ids)
        status_volumes = list(
            filter(lambda v: v.status in valid_status, nosnap_volumes))
        valid_volumes = [v for v in status_volumes
                         if getattr(v, 'migration_status') == 'success' or
                         getattr(v, 'migration_status') is None]
        LOG.info("valid volumes: %s", valid_volumes)

        return valid_volumes

    def group_pools(self, pools, threshold):
        """group volume pools by threshold.

        :param pools: all volume pools
        :param threshold: volume threshold
        :return: under and over threshold pools
        """
        under_pools = list(
            filter(lambda p: float(p.total_capacity_gb) -
                   float(p.free_capacity_gb) <
                   float(p.total_capacity_gb) * threshold, pools))

        over_pools = list(
            filter(lambda p: float(p.total_capacity_gb) -
                   float(p.free_capacity_gb) >=
                   float(p.total_capacity_gb) * threshold, pools))

        return over_pools, under_pools

    def get_volume_type_by_name(self, cinder, backendname):
        # return list of pool type
        if backendname in self.pool_type_cache.keys():
            return self.pool_type_cache.get(backendname)

        volume_type_list = cinder.get_volume_type_list()
        volume_type = list(filter(
            lambda volume_type:
                volume_type.extra_specs.get(
                    'volume_backend_name') == backendname, volume_type_list))
        if volume_type:
            self.pool_type_cache[backendname] = volume_type
            return self.pool_type_cache.get(backendname)
        else:
            return []

    def migrate_fit(self, volume, threshold):
        target_pool_name = None
        if volume.volume_type:
            LOG.info("volume %s type %s", volume.id, volume.volume_type)
            return target_pool_name
        self.dest_pools.sort(
            key=lambda p: float(p.free_capacity_gb) /
            float(p.total_capacity_gb))
        for pool in reversed(self.dest_pools):
            total_cap = float(pool.total_capacity_gb)
            allocated = float(pool.allocated_capacity_gb)
            ratio = pool.max_over_subscription_ratio
            if total_cap * ratio < allocated + float(volume.size):
                LOG.info("pool %s allocated over", pool.name)
                continue
            free_cap = float(pool.free_capacity_gb) - float(volume.size)
            if free_cap > (1 - threshold) * total_cap:
                target_pool_name = pool.name
                index = self.dest_pools.index(pool)
                setattr(self.dest_pools[index], 'free_capacity_gb',
                        str(free_cap))
                LOG.info("volume: get pool %s for vol %s", target_pool_name,
                         volume.name)
                break
        return target_pool_name

    def check_pool_type(self, volume, dest_pool):
        target_type = None
        src_extra_specs = {}
        # check type feature
        if not volume.volume_type:
            return target_type
        volume_type_list = self.cinder.get_volume_type_list()
        volume_type = list(filter(
            lambda volume_type:
                volume_type.name == volume.volume_type, volume_type_list))
        if volume_type:
            src_extra_specs = volume_type[0].extra_specs
            src_extra_specs.pop('volume_backend_name', None)

        backendname = getattr(dest_pool, 'volume_backend_name')
        dst_pool_type = self.get_volume_type_by_name(self.cinder, backendname)

        for src_key in src_extra_specs.keys():
            dst_pool_type = [pt for pt in dst_pool_type
                             if pt.extra_specs.get(src_key) ==
                             src_extra_specs.get(src_key)]
        if dst_pool_type:
            if volume.volume_type:
                if dst_pool_type[0].name != volume.volume_type:
                    target_type = dst_pool_type[0].name
            else:
                target_type = dst_pool_type[0].name
        return target_type

    def retype_fit(self, volume, threshold):
        target_type = None
        self.dest_pools.sort(
            key=lambda p: float(p.free_capacity_gb) /
            float(p.total_capacity_gb))
        for pool in reversed(self.dest_pools):
            backendname = getattr(pool, 'volume_backend_name')
            pool_type = self.get_volume_type_by_name(self.cinder, backendname)
            LOG.info("volume: pool %s, type %s", pool.name, pool_type)
            if pool_type is None:
                continue
            total_cap = float(pool.total_capacity_gb)
            allocated = float(pool.allocated_capacity_gb)
            ratio = pool.max_over_subscription_ratio
            if total_cap * ratio < allocated + float(volume.size):
                LOG.info("pool %s allocated over", pool.name)
                continue
            free_cap = float(pool.free_capacity_gb) - float(volume.size)
            if free_cap > (1 - threshold) * total_cap:
                target_type = self.check_pool_type(volume, pool)
                if target_type is None:
                    continue
                index = self.dest_pools.index(pool)
                setattr(self.dest_pools[index], 'free_capacity_gb',
                        str(free_cap))
                LOG.info("volume: get type %s for vol %s", target_type,
                         volume.name)
                break
        return target_type

    def get_actions(self, pool, volumes, threshold):
        """get volume, pool key-value action

        return: retype, migrate dict
        """
        retype_dicts = dict()
        migrate_dicts = dict()
        total_cap = float(pool.total_capacity_gb)
        used_cap = float(pool.total_capacity_gb) - float(pool.free_capacity_gb)
        seek_flag = True

        volumes_in_pool = list(
            filter(lambda v: getattr(v, 'os-vol-host-attr:host') == pool.name,
                   volumes))
        LOG.info("volumes in pool: %s", str(volumes_in_pool))
        if not volumes_in_pool:
            return retype_dicts, migrate_dicts
        ava_volumes = list(filter(lambda v: v.status == 'available',
                           volumes_in_pool))
        ava_volumes.sort(key=lambda v: float(v.size))
        LOG.info("available volumes in pool: %s ", str(ava_volumes))
        for vol in ava_volumes:
            vol_flag = False
            migrate_pool = self.migrate_fit(vol, threshold)
            if migrate_pool:
                migrate_dicts[vol.id] = migrate_pool
                vol_flag = True
            else:
                target_type = self.retype_fit(vol, threshold)
                if target_type:
                    retype_dicts[vol.id] = target_type
                    vol_flag = True
            if vol_flag:
                used_cap -= float(vol.size)
                if used_cap < threshold * total_cap:
                    seek_flag = False
                    break
        if seek_flag:
            noboot_volumes = list(
                filter(lambda v: v.bootable.lower() == 'false' and
                       v.status == 'in-use', volumes_in_pool))
            noboot_volumes.sort(key=lambda v: float(v.size))
            LOG.info("noboot volumes: %s ", str(noboot_volumes))
            for vol in noboot_volumes:
                vol_flag = False
                migrate_pool = self.migrate_fit(vol, threshold)
                if migrate_pool:
                    migrate_dicts[vol.id] = migrate_pool
                    vol_flag = True
                else:
                    target_type = self.retype_fit(vol, threshold)
                    if target_type:
                        retype_dicts[vol.id] = target_type
                        vol_flag = True
                if vol_flag:
                    used_cap -= float(vol.size)
                    if used_cap < threshold * total_cap:
                        seek_flag = False
                        break

        if seek_flag:
            boot_volumes = list(
                filter(lambda v: v.bootable.lower() == 'true' and
                       v.status == 'in-use', volumes_in_pool)
            )
            boot_volumes.sort(key=lambda v: float(v.size))
            LOG.info("boot volumes: %s ", str(boot_volumes))
            for vol in boot_volumes:
                vol_flag = False
                migrate_pool = self.migrate_fit(vol, threshold)
                if migrate_pool:
                    migrate_dicts[vol.id] = migrate_pool
                    vol_flag = True
                else:
                    target_type = self.retype_fit(vol, threshold)
                    if target_type:
                        retype_dicts[vol.id] = target_type
                        vol_flag = True
                if vol_flag:
                    used_cap -= float(vol.size)
                    if used_cap < threshold * total_cap:
                        seek_flag = False
                        break
        return retype_dicts, migrate_dicts

    def pre_execute(self):
        LOG.info("Initializing " + self.get_display_name() + " Strategy")
        self.volume_threshold = self.input_parameters.volume_threshold

    def do_execute(self, audit=None):
        """Strategy execution phase

        This phase is where you should put the main logic of your strategy.
        """
        all_pools = self.get_pools(self.cinder)
        all_volumes = self.get_volumes(self.cinder)
        threshold = float(self.volume_threshold) / 100
        self.source_pools, self.dest_pools = self.group_pools(
            all_pools, threshold)
        LOG.info(" source pools: %s dest pools:%s",
                 self.source_pools, self.dest_pools)
        if not self.source_pools:
            LOG.info("No pools require optimization")
            return

        if not self.dest_pools:
            LOG.info("No enough pools for optimization")
            return
        for source_pool in self.source_pools:
            retype_actions, migrate_actions = self.get_actions(
                source_pool, all_volumes, threshold)
            for vol_id, pool_type in retype_actions.items():
                vol = [v for v in all_volumes if v.id == vol_id]
                parameters = {'migration_type': 'retype',
                              'destination_type': pool_type,
                              'resource_name': vol[0].name}
                self.solution.add_action(action_type='volume_migrate',
                                         resource_id=vol_id,
                                         input_parameters=parameters)
            for vol_id, pool_name in migrate_actions.items():
                vol = [v for v in all_volumes if v.id == vol_id]
                parameters = {'migration_type': 'migrate',
                              'destination_node': pool_name,
                              'resource_name': vol[0].name}
                self.solution.add_action(action_type='volume_migrate',
                                         resource_id=vol_id,
                                         input_parameters=parameters)

    def post_execute(self):
        """Post-execution phase

        """
        self.solution.set_efficacy_indicators(
            instance_migrations_count=0,
            instances_count=0,
        )



# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from watcher.common import exception
from watcher.decision_engine.loading import default as default_loader
from watcher.decision_engine.strategy.selection import (
    default as default_selector)
from watcher.decision_engine.strategy import strategies
from watcher.tests import base


class TestStrategySelector(base.TestCase):

    @mock.patch.object(default_loader.DefaultStrategyLoader, 'load')
    def test_select_with_strategy_name(self, m_load):
        expected_goal = 'dummy'
        expected_strategy = "dummy"
        strategy_selector = default_selector.DefaultStrategySelector(
            expected_goal, expected_strategy, osc=None)
        strategy_selector.select()
        m_load.assert_called_once_with(expected_strategy, osc=None)

    @mock.patch.object(default_loader.DefaultStrategyLoader, 'load')
    @mock.patch.object(default_loader.DefaultStrategyLoader, 'list_available')
    def test_select_with_goal_name_only(self, m_list_available, m_load):
        m_list_available.return_value = {"dummy": strategies.DummyStrategy}
        expected_goal = 'dummy'
        expected_strategy = "dummy"
        strategy_selector = default_selector.DefaultStrategySelector(
            expected_goal, osc=None)
        strategy_selector.select()
        m_load.assert_called_once_with(expected_strategy, osc=None)

    def test_select_non_existing_strategy(self):
        strategy_selector = default_selector.DefaultStrategySelector(
            "dummy", "NOT_FOUND")
        self.assertRaises(exception.LoadingError, strategy_selector.select)

    @mock.patch.object(default_loader.DefaultStrategyLoader, 'list_available')
    def test_select_no_available_strategy_for_goal(self, m_list_available):
        m_list_available.return_value = {}
        strategy_selector = default_selector.DefaultStrategySelector("dummy")
        self.assertRaises(exception.NoAvailableStrategyForGoal,
                          strategy_selector.select)



# -*- encoding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import abc

from oslo_config import cfg
from oslo_log import log

from watcher.common import clients
from watcher.common import exception
from watcher.common import nova_helper
from watcher.common import utils
from watcher.decision_engine.planner import base
from watcher import objects

LOG = log.getLogger(__name__)


class WorkloadStabilizationPlanner(base.BasePlanner):
    """Workload Stabilization planner implementation

    This implementation comes with basic rules with a set of action types that
    are weighted. An action having a lower weight will be scheduled before the
    other ones. The set of action types can be specified by 'weights' in the
    ``watcher.conf``. You need to associate a different weight to all available
    actions into the configuration file, otherwise you will get an error when
    the new action will be referenced in the solution produced by a strategy.

    *Limitations*

    - This is a proof of concept that is not meant to be used in production
    """

    def __init__(self, config):
        super(WorkloadStabilizationPlanner, self).__init__(config)
        self._osc = clients.OpenStackClients()

    @property
    def osc(self):
        return self._osc

    weights_dict = {
        'turn_host_to_acpi_s3_state': 0,
        'resize': 1,
        'migrate': 2,
        'sleep': 3,
        'change_nova_service_state': 4,
        'nop': 5,
    }

    @classmethod
    def get_config_opts(cls):
        return [
            cfg.DictOpt(
                'weights',
                help="These weights are used to schedule the actions",
                default=cls.weights_dict),
        ]

    def create_action(self,
                      action_plan_id,
                      action_type,
                      input_parameters=None):
        uuid = utils.generate_uuid()
        action = {
            'uuid': uuid,
            'action_plan_id': int(action_plan_id),
            'action_type': action_type,
            'input_parameters': input_parameters,
            'state': objects.action.State.PENDING,
            'parents': None
        }

        return action

    def load_child_class(self, child_name):
        for c in BaseActionValidator.__subclasses__():
            if child_name == c.action_name:
                return c()
        return None

    def schedule(self, context, audit_id, solution):
        LOG.debug('Creating an action plan for the audit uuid: %s', audit_id)
        weights = self.config.weights
        action_plan = self._create_action_plan(context, audit_id, solution)

        actions = list(solution.actions)
        to_schedule = []
        for action in actions:
            json_action = self.create_action(
                action_plan_id=action_plan.id,
                action_type=action.get('action_type'),
                input_parameters=action.get('input_parameters'))
            to_schedule.append((weights[action.get('action_type')],
                                json_action))

        self._create_efficacy_indicators(
            context, action_plan.id, solution.efficacy_indicators)

        # scheduling
        scheduled = sorted(to_schedule, key=lambda weight: (weight[0]),
                           reverse=True)
        if len(scheduled) == 0:
            LOG.warning("The action plan is empty")
            action_plan.state = objects.action_plan.State.SUCCEEDED
            action_plan.save()
        else:
            resource_action_map = {}
            scheduled_actions = [x[1] for x in scheduled]
            for action in scheduled_actions:
                a_type = action['action_type']
                if a_type != 'turn_host_to_acpi_s3_state':
                    plugin_action = self.load_child_class(
                        action.get("action_type"))
                    if not plugin_action:
                        raise exception.UnsupportedActionType(
                            action_type=action.get("action_type"))
                    db_action = self._create_action(context, action)
                    parents = plugin_action.validate_parents(
                        resource_action_map, action)
                    if parents:
                        db_action.parents = parents
                        db_action.save()
                # if we have an action that will make host unreachable, we need
                # to complete all actions (resize and migration type)
                # related to the host.
                # Note(alexchadin): turn_host_to_acpi_s3_state doesn't
                # actually exist. Placed code shows relations between
                # action types.
                # TODO(alexchadin): add turn_host_to_acpi_s3_state action type.
                else:
                    host_to_acpi_s3 = action['input_parameters']['resource_id']
                    host_actions = resource_action_map.get(host_to_acpi_s3)
                    action_parents = []
                    if host_actions:
                        resize_actions = [x[0] for x in host_actions
                                          if x[1] == 'resize']
                        migrate_actions = [x[0] for x in host_actions
                                           if x[1] == 'migrate']
                        resize_migration_parents = [
                            x.parents for x in
                            [objects.Action.get_by_uuid(context, resize_action)
                             for resize_action in resize_actions]]
                        # resize_migration_parents should be one level list
                        resize_migration_parents = [
                            parent for sublist in resize_migration_parents
                            for parent in sublist]
                        action_parents.extend([uuid for uuid in
                                               resize_actions])
                        action_parents.extend([uuid for uuid in
                                              migrate_actions if uuid not in
                                              resize_migration_parents])
                    db_action = self._create_action(context, action)
                    db_action.parents = action_parents
                    db_action.save()

        return action_plan

    def _create_action_plan(self, context, audit_id, solution):
        strategy = objects.Strategy.get_by_name(
            context, solution.strategy.name)

        action_plan_dict = {
            'uuid': utils.generate_uuid(),
            'audit_id': audit_id,
            'strategy_id': strategy.id,
            'state': objects.action_plan.State.RECOMMENDED,
            'global_efficacy': solution.global_efficacy,
        }

        new_action_plan = objects.ActionPlan(context, **action_plan_dict)
        new_action_plan.create()

        return new_action_plan

    def _create_efficacy_indicators(self, context, action_plan_id, indicators):
        efficacy_indicators = []
        for indicator in indicators:
            efficacy_indicator_dict = {
                'uuid': utils.generate_uuid(),
                'name': indicator.name,
                'description': indicator.description,
                'unit': indicator.unit,
                'value': indicator.value,
                'action_plan_id': action_plan_id,
            }
            new_efficacy_indicator = objects.EfficacyIndicator(
                context, **efficacy_indicator_dict)
            new_efficacy_indicator.create()

            efficacy_indicators.append(new_efficacy_indicator)
        return efficacy_indicators

    def _create_action(self, context, _action):
        try:
            LOG.debug("Creating the %s in the Watcher database",
                      _action.get("action_type"))

            new_action = objects.Action(context, **_action)
            new_action.create()

            return new_action
        except Exception as exc:
            LOG.exception(exc)
            raise


class BaseActionValidator(object):
    action_name = None

    def __init__(self):
        super(BaseActionValidator, self).__init__()
        self._osc = None

    @property
    def osc(self):
        if not self._osc:
            self._osc = clients.OpenStackClients()
        return self._osc

    @abc.abstractmethod
    def validate_parents(self, resource_action_map, action):
        raise NotImplementedError()

    def _mapping(self, resource_action_map, resource_id, action_uuid,
                 action_type):
        if resource_id not in resource_action_map:
            resource_action_map[resource_id] = [(action_uuid,
                                                 action_type,)]
        else:
            resource_action_map[resource_id].append((action_uuid,
                                                     action_type,))


class MigrationActionValidator(BaseActionValidator):
    action_name = "migrate"

    def validate_parents(self, resource_action_map, action):
        instance_uuid = action['input_parameters']['resource_id']
        host_name = action['input_parameters']['source_node']
        self._mapping(resource_action_map, instance_uuid, action['uuid'],
                      'migrate')
        self._mapping(resource_action_map, host_name, action['uuid'],
                      'migrate')


class ResizeActionValidator(BaseActionValidator):
    action_name = "resize"

    def validate_parents(self, resource_action_map, action):
        nova = nova_helper.NovaHelper(osc=self.osc)
        instance_uuid = action['input_parameters']['resource_id']
        parent_actions = resource_action_map.get(instance_uuid)
        host_of_instance = nova.get_hostname(
            nova.get_instance_by_uuid(instance_uuid)[0])
        self._mapping(resource_action_map, host_of_instance, action['uuid'],
                      'resize')
        if parent_actions:
            return [x[0] for x in parent_actions]
        else:
            return []


class ChangeNovaServiceStateActionValidator(BaseActionValidator):
    action_name = "change_nova_service_state"

    def validate_parents(self, resource_action_map, action):
        host_name = action['input_parameters']['resource_id']
        self._mapping(resource_action_map, host_name, action['uuid'],
                      'change_nova_service_state')
        return []


class SleepActionValidator(BaseActionValidator):
    action_name = "sleep"

    def validate_parents(self, resource_action_map, action):
        return []


class NOPActionValidator(BaseActionValidator):
    action_name = "nop"

    def validate_parents(self, resource_action_map, action):
        return []



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 b<>com
#
# Authors: Vincent FRANCOISE <Vincent.FRANCOISE@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os_resource_classes as orc
from oslo_log import log
from watcher.common import exception
from watcher.common import nova_helper
from watcher.common import placement_helper
from watcher.common import utils
from watcher.decision_engine.model import element
from watcher.decision_engine.model.notification import base
from watcher.decision_engine.model.notification import filtering

LOG = log.getLogger(__name__)


class NovaNotification(base.NotificationEndpoint):

    def __init__(self, collector):
        super(NovaNotification, self).__init__(collector)
        self._nova = None
        self._placement_helper = None

    @property
    def nova(self):
        if self._nova is None:
            self._nova = nova_helper.NovaHelper()
        return self._nova

    @property
    def placement_helper(self):
        if self._placement_helper is None:
            self._placement_helper = placement_helper.PlacementHelper()
        return self._placement_helper

    def get_or_create_instance(self, instance_uuid, node_name=None):
        try:
            node = None
            if node_name:
                node = self.get_or_create_node(node_name)
        except exception.ComputeNodeNotFound:
            LOG.warning("Could not find compute node %(node)s for "
                        "instance %(instance)s",
                        dict(node=node_name, instance=instance_uuid))
        try:
            instance = self.cluster_data_model.get_instance_by_uuid(
                instance_uuid)
        except exception.InstanceNotFound:
            # The instance didn't exist yet so we create a new instance object
            LOG.debug("New instance created: %s", instance_uuid)
            instance = element.Instance(uuid=instance_uuid)

            self.cluster_data_model.add_instance(instance)
            if node:
                self.cluster_data_model.map_instance(instance, node)

        return instance

    def update_instance(self, instance, data):
        n_version = float(data['nova_object.version'])
        instance_data = data['nova_object.data']
        instance_flavor_data = instance_data['flavor']['nova_object.data']

        memory_mb = instance_flavor_data['memory_mb']
        num_cores = instance_flavor_data['vcpus']
        disk_gb = instance_flavor_data['root_gb']
        instance_metadata = data['nova_object.data']['metadata']

        instance.update({
            'state': instance_data['state'],
            'hostname': instance_data['host_name'],
            # this is the user-provided display name of the server which is not
            # guaranteed to be unique nor is it immutable.
            'name': instance_data['display_name'],
            'memory': memory_mb,
            'vcpus': num_cores,
            'disk': disk_gb,
            'metadata': instance_metadata,
            'project_id': instance_data['tenant_id']
        })
        # locked was added in nova notification payload version 1.1
        if n_version > 1.0:
            instance.update({'locked': instance_data['locked']})

        try:
            node = self.get_or_create_node(instance_data['host'])
        except exception.ComputeNodeNotFound as exc:
            LOG.exception(exc)
            # If we can't create the node, we consider the instance as unmapped
            node = None

        self.update_instance_mapping(instance, node)

    def update_compute_node(self, node, data):
        """Update the compute node using the notification data."""
        node_data = data['nova_object.data']
        node_state = (
            element.ServiceState.OFFLINE.value
            if node_data['forced_down'] else element.ServiceState.ONLINE.value)
        node_status = (
            element.ServiceState.DISABLED.value
            if node_data['disabled'] else element.ServiceState.ENABLED.value)
        disabled_reason = (
            node_data['disabled_reason']
            if node_data['disabled'] else None)

        node.update({
            'hostname': node_data['host'],
            'state': node_state,
            'status': node_status,
            'disabled_reason': disabled_reason,
        })

    def create_compute_node(self, uuid_or_name):
        """Create the computeNode node."""
        try:
            if utils.is_uuid_like(uuid_or_name):
                _node = self.nova.get_compute_node_by_uuid(uuid_or_name)
            else:
                _node = self.nova.get_compute_node_by_hostname(uuid_or_name)
            inventories = self.placement_helper.get_inventories(_node.id)
            if inventories and orc.VCPU in inventories:
                vcpus = inventories[orc.VCPU]['total']
                vcpu_reserved = inventories[orc.VCPU]['reserved']
                vcpu_ratio = inventories[orc.VCPU]['allocation_ratio']
            else:
                vcpus = _node.vcpus
                vcpu_reserved = 0
                vcpu_ratio = 1.0

            if inventories and orc.MEMORY_MB in inventories:
                memory_mb = inventories[orc.MEMORY_MB]['total']
                memory_mb_reserved = inventories[orc.MEMORY_MB]['reserved']
                memory_ratio = inventories[orc.MEMORY_MB]['allocation_ratio']
            else:
                memory_mb = _node.memory_mb
                memory_mb_reserved = 0
                memory_ratio = 1.0

            # NOTE(licanwei): A BP support-shared-storage-resource-provider
            # will move DISK_GB from compute node to shared storage RP.
            # Here may need to be updated when the nova BP released.
            if inventories and orc.DISK_GB in inventories:
                disk_capacity = inventories[orc.DISK_GB]['total']
                disk_gb_reserved = inventories[orc.DISK_GB]['reserved']
                disk_ratio = inventories[orc.DISK_GB]['allocation_ratio']
            else:
                disk_capacity = _node.local_gb
                disk_gb_reserved = 0
                disk_ratio = 1.0

            # build up the compute node.
            node_attributes = {
                # The id of the hypervisor as a UUID from version 2.53.
                "uuid": _node.id,
                "hostname": _node.service["host"],
                "memory": memory_mb,
                "memory_ratio": memory_ratio,
                "memory_mb_reserved": memory_mb_reserved,
                "disk": disk_capacity,
                "disk_gb_reserved": disk_gb_reserved,
                "disk_ratio": disk_ratio,
                "vcpus": vcpus,
                "vcpu_reserved": vcpu_reserved,
                "vcpu_ratio": vcpu_ratio,
                "state": _node.state,
                "status": _node.status,
                "disabled_reason": _node.service["disabled_reason"]}

            node = element.ComputeNode(**node_attributes)
            self.cluster_data_model.add_node(node)
            LOG.debug("New compute node mapped: %s", node.uuid)
            return node
        except Exception as exc:
            LOG.exception(exc)
            LOG.debug("Could not refresh the node %s.", uuid_or_name)
            raise exception.ComputeNodeNotFound(name=uuid_or_name)

    def get_or_create_node(self, uuid_or_name):
        if uuid_or_name is None:
            LOG.debug("Compute node UUID or name not provided: skipping")
            return
        try:
            if utils.is_uuid_like(uuid_or_name):
                return self.cluster_data_model.get_node_by_uuid(uuid_or_name)
            else:
                return self.cluster_data_model.get_node_by_name(uuid_or_name)
        except exception.ComputeNodeNotFound:
            # The node didn't exist yet so we create a new node object
            node = self.create_compute_node(uuid_or_name)
            LOG.debug("New compute node created: %s", uuid_or_name)
            return node

    def update_instance_mapping(self, instance, node):
        if node is None:
            self.cluster_data_model.add_instance(instance)
            LOG.debug("Instance %s not yet attached to any node: skipping",
                      instance.uuid)
            return
        try:
            try:
                current_node = (
                    self.cluster_data_model.get_node_by_instance_uuid(
                        instance.uuid))
            except exception.ComputeResourceNotFound as exc:
                LOG.exception(exc)
                # If we can't create the node,
                # we consider the instance as unmapped
                current_node = None

            LOG.debug("Mapped node %s found", node.uuid)
            if current_node and node != current_node:
                LOG.debug("Unmapping instance %s from %s",
                          instance.uuid, node.uuid)
                self.cluster_data_model.unmap_instance(instance, current_node)
        except exception.InstanceNotFound:
            # The instance didn't exist yet so we map it for the first time
            LOG.debug("New instance: mapping it to %s", node.uuid)
        finally:
            if node:
                self.cluster_data_model.map_instance(instance, node)
                LOG.debug("Mapped instance %s to %s", instance.uuid, node.uuid)

    def delete_instance(self, instance, node):
        try:
            self.cluster_data_model.delete_instance(instance, node)
        except Exception:
            LOG.info("Instance %s already deleted", instance.uuid)

    def delete_node(self, node):
        try:
            self.cluster_data_model.remove_node(node)
        except Exception:
            LOG.info("Node %s already deleted", node.uuid)


class VersionedNotification(NovaNotification):
    publisher_id_regex = r'^nova-.*'

    def service_updated(self, payload):
        node_data = payload['nova_object.data']
        node_name = node_data['host']
        try:
            node = self.get_or_create_node(node_name)
            self.update_compute_node(node, payload)
        except exception.ComputeNodeNotFound as exc:
            LOG.exception(exc)

    def service_deleted(self, payload):
        node_data = payload['nova_object.data']
        node_name = node_data['host']
        try:
            node = self.get_or_create_node(node_name)
            self.delete_node(node)
        except exception.ComputeNodeNotFound as exc:
            LOG.exception(exc)

    def instance_updated(self, payload):
        instance_data = payload['nova_object.data']
        instance_uuid = instance_data['uuid']
        instance_state = instance_data['state']
        node_name = instance_data.get('host')
        # if instance state is building, don't update data model
        if instance_state == 'building':
            return

        instance = self.get_or_create_instance(instance_uuid, node_name)

        self.update_instance(instance, payload)

    def instance_created(self, payload):
        instance_data = payload['nova_object.data']
        instance_uuid = instance_data['uuid']
        instance = element.Instance(uuid=instance_uuid)
        self.cluster_data_model.add_instance(instance)

        node_name = instance_data.get('host')
        if node_name:
            node = self.get_or_create_node(node_name)
            self.cluster_data_model.map_instance(instance, node)

        self.update_instance(instance, payload)

    def instance_deleted(self, payload):
        instance_data = payload['nova_object.data']
        instance_uuid = instance_data['uuid']
        node_name = instance_data.get('host')
        instance = self.get_or_create_instance(instance_uuid, node_name)

        try:
            node = self.get_or_create_node(instance_data['host'])
        except exception.ComputeNodeNotFound as exc:
            LOG.exception(exc)
            # If we can't create the node, we consider the instance as unmapped
            node = None

        self.delete_instance(instance, node)

    notification_mapping = {
        'instance.create.end': instance_created,
        'instance.lock': instance_updated,
        'instance.unlock': instance_updated,
        'instance.pause.end': instance_updated,
        'instance.power_off.end': instance_updated,
        'instance.power_on.end': instance_updated,
        'instance.resize_confirm.end': instance_updated,
        'instance.restore.end': instance_updated,
        'instance.resume.end': instance_updated,
        'instance.shelve.end': instance_updated,
        'instance.shutdown.end': instance_updated,
        'instance.suspend.end': instance_updated,
        'instance.unpause.end': instance_updated,
        'instance.unrescue.end': instance_updated,
        'instance.unshelve.end': instance_updated,
        'instance.rebuild.end': instance_updated,
        'instance.rescue.end': instance_updated,
        'instance.update': instance_updated,
        'instance.live_migration_force_complete.end': instance_updated,
        'instance.live_migration_post.end': instance_updated,
        'instance.delete.end': instance_deleted,
        'instance.soft_delete.end': instance_deleted,
        'service.create': service_updated,
        'service.delete': service_deleted,
        'service.update': service_updated,
        }

    @property
    def filter_rule(self):
        """Nova notification filter"""
        return filtering.NotificationFilter(
            publisher_id=self.publisher_id_regex,
        )

    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        LOG.info("Event '%(event)s' received from %(publisher)s "
                 "with metadata %(metadata)s",
                 dict(event=event_type,
                      publisher=publisher_id,
                      metadata=metadata))
        func = self.notification_mapping.get(event_type)
        if func:
            # The nova CDM is not built until an audit is performed.
            if self.cluster_data_model:
                LOG.debug(payload)
                func(self, payload)
            else:
                LOG.debug('Nova CDM has not yet been built; ignoring '
                          'notifications until an audit is performed.')



# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
#
# Authors: Jean-Emile DARTOIS <jean-emile.dartois@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
The :ref:`Watcher Planner <watcher_planner_definition>` is part of the
:ref:`Watcher Decision Engine <watcher_decision_engine_definition>`.

This module takes the set of :ref:`Actions <action_definition>` generated by a
:ref:`Strategy <strategy_definition>` and builds the design of a workflow which
defines how-to schedule in time those different
:ref:`Actions <action_definition>` and for each
:ref:`Action <action_definition>` what are the prerequisite conditions.

It is important to schedule :ref:`Actions <action_definition>` in time in order
to prevent overload of the :ref:`Cluster <cluster_definition>` while applying
the :ref:`Action Plan <action_plan_definition>`. For example, it is important
not to migrate too many instances at the same time in order to avoid a network
congestion which may decrease the :ref:`SLA <sla_definition>` for
:ref:`Customers <customer_definition>`.

It is also important to schedule :ref:`Actions <action_definition>` in order to
avoid security issues such as denial of service on core OpenStack services.

:ref:`Some default implementations are provided <watcher_planners>`, but it is
possible to :ref:`develop new implementations <implement_planner_plugin>`
which are dynamically loaded by Watcher at launch time.

See :doc:`../architecture` for more details on this component.
"""

import abc

from watcher.common.loader import loadable


class BasePlanner(loadable.Loadable, metaclass=abc.ABCMeta):

    @classmethod
    def get_config_opts(cls):
        """Defines the configuration options to be associated to this loadable

        :return: A list of configuration options relative to this Loadable
        :rtype: list of :class:`oslo_config.cfg.Opt` instances
        """
        return []

    @abc.abstractmethod
    def schedule(self, context, audit_uuid, solution):
        """The planner receives a solution to schedule

        :param solution: A solution provided by a strategy for scheduling
        :type solution: :py:class:`~.BaseSolution` subclass instance
        :param audit_uuid: the audit uuid
        :type audit_uuid: str
        :return: Action plan with an ordered sequence of actions such that all
                 security, dependency, and performance requirements are met.
        :rtype: :py:class:`watcher.objects.ActionPlan` instance
        """
        # example: directed acyclic graph
        raise NotImplementedError()



# Copyright 2019 ZTE corporation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from http import HTTPStatus
from oslo_serialization import jsonutils

from watcher.decision_engine import rpcapi as deapi
from watcher.tests.api import base as api_base


class TestListDataModel(api_base.FunctionalTest):

    def setUp(self):
        super(TestListDataModel, self).setUp()
        p_dcapi = mock.patch.object(deapi, 'DecisionEngineAPI')
        self.mock_dcapi = p_dcapi.start()
        self.mock_dcapi().get_data_model_info.return_value = \
            'fake_response_value'
        self.addCleanup(p_dcapi.stop)

    def test_get_all(self):
        response = self.get_json(
            '/data_model/?data_model_type=compute',
            headers={'OpenStack-API-Version': 'infra-optim 1.3'})
        self.assertEqual('fake_response_value', response)

    def test_get_all_not_acceptable(self):
        response = self.get_json(
            '/data_model/?data_model_type=compute',
            headers={'OpenStack-API-Version': 'infra-optim 1.2'},
            expect_errors=True)
        self.assertEqual(HTTPStatus.NOT_ACCEPTABLE, response.status_int)


class TestDataModelPolicyEnforcement(api_base.FunctionalTest):

    def setUp(self):
        super(TestDataModelPolicyEnforcement, self).setUp()
        p_dcapi = mock.patch.object(deapi, 'DecisionEngineAPI')
        self.mock_dcapi = p_dcapi.start()
        self.addCleanup(p_dcapi.stop)

    def _common_policy_check(self, rule, func, *arg, **kwarg):
        self.policy.set_rules({
            "admin_api": "(role:admin or role:administrator)",
            "default": "rule:admin_api",
            rule: "rule:defaut"})
        response = func(*arg, **kwarg)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(
            "Policy doesn't allow %s to be performed." % rule,
            jsonutils.loads(response.json['error_message'])['faultstring'])

    def test_policy_disallow_get_all(self):
        self._common_policy_check(
            "data_model:get_all", self.get_json,
            "/data_model/?data_model_type=compute",
            headers={'OpenStack-API-Version': 'infra-optim 1.3'},
            expect_errors=True)


class TestDataModelEnforcementWithAdminContext(
        TestListDataModel, api_base.AdminRoleTest):

    def setUp(self):
        super(TestDataModelEnforcementWithAdminContext, self).setUp()
        self.policy.set_rules({
            "admin_api": "(role:admin or role:administrator)",
            "default": "rule:admin_api",
            "data_model:get_all": "rule:default"})



# -*- encoding: utf-8 -*-
#
# Copyright © 2012 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from http import HTTPStatus
from oslo_config import cfg
from pecan import hooks

from watcher.common import context


class ContextHook(hooks.PecanHook):
    """Configures a request context and attaches it to the request.

    The following HTTP request headers are used:

    X-User:
        Used for context.user.

    X-User-Id:
        Used for context.user_id.

    X-Project-Name:
        Used for context.project.

    X-Project-Id:
        Used for context.project_id.

    X-Auth-Token:
        Used for context.auth_token.

    """

    def before(self, state):
        headers = state.request.headers
        user = headers.get('X-User')
        user_id = headers.get('X-User-Id')
        project = headers.get('X-Project-Name')
        project_id = headers.get('X-Project-Id')
        domain_id = headers.get('X-User-Domain-Id')
        domain_name = headers.get('X-User-Domain-Name')
        auth_token = headers.get('X-Storage-Token')
        auth_token = headers.get('X-Auth-Token', auth_token)
        show_deleted = headers.get('X-Show-Deleted')
        auth_token_info = state.request.environ.get('keystone.token_info')
        roles = (headers.get('X-Roles', None) and
                 headers.get('X-Roles').split(','))

        state.request.context = context.make_context(
            auth_token=auth_token,
            auth_token_info=auth_token_info,
            user=user,
            user_id=user_id,
            project=project,
            project_id=project_id,
            domain_id=domain_id,
            domain_name=domain_name,
            show_deleted=show_deleted,
            roles=roles)


class NoExceptionTracebackHook(hooks.PecanHook):
    """Workaround rpc.common: deserialize_remote_exception.

    deserialize_remote_exception builds rpc exception traceback into error
    message which is then sent to the client. Such behavior is a security
    concern so this hook is aimed to cut-off traceback from the error message.
    """
    # NOTE(max_lobur): 'after' hook used instead of 'on_error' because
    # 'on_error' never fired for wsme+pecan pair. wsme @wsexpose decorator
    # catches and handles all the errors, so 'on_error' dedicated for unhandled
    # exceptions never fired.
    def after(self, state):
        # Omit empty body. Some errors may not have body at this level yet.
        if not state.response.body:
            return

        # Do nothing if there is no error.
        # Status codes in the range 200 (OK) to 399 (400 = BAD_REQUEST) are not
        # an error.
        if (HTTPStatus.OK <= state.response.status_int <
                HTTPStatus.BAD_REQUEST):
            return

        json_body = state.response.json
        # Do not remove traceback when traceback config is set
        if cfg.CONF.debug:
            return

        faultstring = json_body.get('faultstring')
        traceback_marker = 'Traceback (most recent call last):'
        if faultstring and traceback_marker in faultstring:
            # Cut-off traceback.
            faultstring = faultstring.split(traceback_marker, 1)[0]
            # Remove trailing newlines and spaces if any.
            json_body['faultstring'] = faultstring.rstrip()
            # Replace the whole json. Cannot change original one because it's
            # generated on the fly.
            state.response.json = json_body
